import collections
import importlib
import inspect
import types
from abjad.system.AbjadValueObject import AbjadValueObject


class StorageFormatManager(AbjadValueObject):
    """
    Manages Abjad object storage formats.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Storage formatting'

    __slots__ = (
        '_client',
        '_format_specification',
        '_signature_accepts_args',
        '_signature_accepts_kwargs',
        '_signature_keyword_names',
        '_signature_positional_names',
        )

    _exclude_tools_package = (
        'core',
        'indicators',
        'markup',
        'mathtools',
        'meter',
        'pitch',
        'scheme',
        'timespans',
        'utilities',
        )

    _unindented_whitespace = '', '', ', '

    _indented_whitespace = '    ', '\n', ',\n'

    ### INITIALIZER ###

    def __init__(self, client=None):
        self._client = client
        self._format_specification = None
        (
            self._signature_positional_names,
            self._signature_keyword_names,
            self._signature_accepts_args,
            self._signature_accepts_kwargs,
            ) = self.inspect_signature(self._client)

    ### PRIVATE METHODS ###

    def _dispatch_formatting(
        self,
        as_storage_format=True,
        is_indented=True,
        ):
        import abjad
        if isinstance(self._client, types.MethodType):
            return self._format_method()
        elif isinstance(self._client, type):
            return self._format_class(as_storage_format, is_indented)
        elif as_storage_format and (
            hasattr(self._client, '_get_storage_format_specification') or
            hasattr(self._client, '_get_format_specification')
            ):
            pieces = self._format_specced_object(
                as_storage_format=as_storage_format,
                )
            return list(pieces)
        elif (
            not as_storage_format and
            hasattr(self._client, '_get_format_specification')
            ):
            pieces = self._format_specced_object(
                as_storage_format=as_storage_format,
                )
            return list(pieces)
        elif isinstance(self._client, (list, tuple)):
            return self._format_sequence(as_storage_format, is_indented)
        elif isinstance(self._client, (
            abjad.OrderedDict,
            collections.OrderedDict,
            )):
            return self._format_ordered_mapping(as_storage_format, is_indented)
        elif isinstance(self._client, dict):
            return self._format_mapping(as_storage_format, is_indented)
        elif isinstance(self._client, float):
            return repr(round(self._client, 15)).split('\n')
        return repr(self._client).split('\n')

    def _format_class(self, as_storage_format, is_indented):
        if as_storage_format:
            root_package_name = self.get_root_package_name()
            root_package = importlib.import_module(root_package_name)
            parts = [root_package_name]
            if self._client.__name__ not in dir(root_package):
                tools_package_name = self.get_tools_package_name()
                parts.append(tools_package_name)
            parts.append(self._client.__name__)
            result = '.'.join(parts)
        else:
            result = self._client.__name__
        return [result]

    def _format_mapping(self, as_storage_format, is_indented):
        result = []
        prefix, infix, suffix = \
            self._get_whitespace(is_indented)
        result.append('{' + infix)
        for key, value in sorted(self._client.items()):
            key_agent = type(self)(key)
            key_pieces = key_agent._dispatch_formatting(
                as_storage_format=as_storage_format,
                is_indented=is_indented,
                )
            value_agent = type(self)(value)
            value_pieces = value_agent._dispatch_formatting(
                as_storage_format=as_storage_format,
                is_indented=is_indented,
                )
            for line in key_pieces[:-1]:
                result.append(prefix + line)
            result.append('{}{}: {}'.format(
                prefix, key_pieces[-1], value_pieces[0]))
            for line in value_pieces[1:]:
                result.append(prefix + line)
            result[-1] = result[-1] + suffix
        if not is_indented:
            result[-1] = result[-1].rstrip(suffix) + infix
        result.append(prefix + '}')
        return result

    def _format_method(self, as_storage_format, is_indented):
        return []

    def _format_ordered_mapping(self, as_storage_format, is_indented):
        result = []
        prefix, infix, suffix = \
            self._get_whitespace(is_indented)
        result.append('[' + infix)
        for item in list(self._client.items()):
            item_agent = type(self)(item)
            item_pieces = item_agent._dispatch_formatting(
                as_storage_format=as_storage_format,
                is_indented=is_indented,
                )
            for line in item_pieces:
                result.append(prefix + line)
            result[-1] = result[-1] + suffix
        if not is_indented:
            result[-1] = result[-1].rstrip(suffix) + infix
        result.append(prefix + ']')
        return result

    def _format_sequence(self, as_storage_format, is_indented):
        result = []
        prefix, infix, suffix = self._get_whitespace(is_indented)
        # just return the repr, if all contents are builtin types
        prototype = (bool, int, float, str, type(None))
        if all(isinstance(x, prototype) for x in self._client):
            piece = repr(self._client)
            if len(piece) < 50:
                return [repr(self._client)]
        if isinstance(self._client, list):
            braces = '[', ']'
        else:
            braces = '(', ')'
        result.append(braces[0] + infix)
        for x in self._client:
            agent = type(self)(x)
            pieces = agent._dispatch_formatting(
                as_storage_format=as_storage_format,
                is_indented=is_indented,
                )
            for line in pieces[:-1]:
                result.append(prefix + line)
            result.append(prefix + pieces[-1] + suffix)
        if not is_indented:
            if isinstance(self._client, list) or 1 < len(self._client):
                result[-1] = result[-1].rstrip(suffix)
            else:
                result[-1] = result[-1].rstrip()
        result.append(prefix + braces[1])
        return result

    def _format_specced_object(self, as_storage_format=True):
        if hasattr(self._client, '_get_format_specification'):
            specification = self._client._get_format_specification()
            if specification.storage_format_forced_override is not None:
                return [specification.storage_format_forced_override]
        formatting_keywords = self._get_formatting_keywords(as_storage_format)
        args_values = formatting_keywords['args_values']
        as_storage_format = formatting_keywords['as_storage_format']
        is_bracketed = formatting_keywords['is_bracketed']
        is_indented = formatting_keywords['is_indented']
        kwargs_names = formatting_keywords['kwargs_names']
        text = formatting_keywords['text']
        result = []
        if is_bracketed:
            result.append('<')
        if text is not None:
            result.append(text)
        else:
            prefix, infix, suffix = self._get_whitespace(is_indented)
            class_name_prefix = self.get_class_name_prefix(as_storage_format)
            positional_argument_pieces = []
            for value in args_values:
                agent = type(self)(value)
                pieces = agent._dispatch_formatting(
                    as_storage_format=as_storage_format,
                    is_indented=is_indented,
                    )
                for piece in pieces[:-1]:
                    positional_argument_pieces.append(prefix + piece)
                positional_argument_pieces.append(prefix + pieces[-1] + suffix)
            keyword_argument_pieces = []
            for name in kwargs_names:
                value = self._get(name)
                if value is None or isinstance(value, types.MethodType):
                    continue
                agent = type(self)(value)
                pieces = agent._dispatch_formatting(
                    as_storage_format=as_storage_format,
                    is_indented=is_indented,
                    )
                pieces[0] = '{}={}'.format(name, pieces[0])
                for piece in pieces[:-1]:
                    keyword_argument_pieces.append(prefix + piece)
                keyword_argument_pieces.append(prefix + pieces[-1] + suffix)
            if not positional_argument_pieces and not keyword_argument_pieces:
                result.append('{}()'.format(class_name_prefix))
            else:
                result.append('{}({}'.format(
                    class_name_prefix,
                    infix,
                    ))
                result.extend(positional_argument_pieces)
                if positional_argument_pieces and not keyword_argument_pieces:
                    result[-1] = result[-1].rstrip(suffix) + infix
                else:
                    result.extend(keyword_argument_pieces)
                if not as_storage_format:
                    result[-1] = result[-1].rstrip(suffix) + infix
                if is_indented:
                    result.append('{})'.format(prefix))
                else:
                    result.append(')')
        if is_bracketed:
            result.append('>')
        if not is_indented:
            return [''.join(result)]
        return result

    def _get(self, name):
        value = None
        try:
            value = getattr(self._client, name, None)
            if value is None:
                value = getattr(self._client, '_' + name, None)
            if value is None:
                value = getattr(self._client, '_' + name.rstrip('_'), None)
        except AttributeError:
            try:
                value = self._client[name]
            except (TypeError, KeyError):
                value = None
        return value

    def _get_formatting_keywords(self, as_storage_format=True):
        # NOTE: This acts to abstract-away our competing spec-specs.
        #       It can probably be removed/reduced in the near future.
        if as_storage_format:
            spec = None
            if hasattr(self._client, '_get_storage_format_specification'):
                spec = self._client._get_storage_format_specification()
            if spec:
                # print('STORAGE', type(self._client), getattr(self._client, 'name', None))
                via = '_get_storage_format_specification()'
                args_values = spec.positional_argument_values
                is_bracketed = False
                is_indented = spec.is_indented
                kwargs_names = spec.keyword_argument_names
                text = spec.storage_format_text
            else:
                spec = self.format_specification
                via = '_get_format_specification()'
                args_values = spec.storage_format_args_values
                is_bracketed = spec.storage_format_is_bracketed
                is_indented = spec.storage_format_is_indented
                kwargs_names = spec.storage_format_kwargs_names
                text = spec.storage_format_text
        else:
            spec = self.format_specification
            via = '_get_format_specification()'
            args_values = spec.repr_args_values
            if args_values is None:
                args_values = spec.storage_format_args_values
            is_bracketed = spec.repr_is_bracketed
            is_indented = spec.repr_is_indented
            kwargs_names = spec.repr_kwargs_names
            if kwargs_names is None:
                kwargs_names = spec.storage_format_kwargs_names
            text = spec.repr_text
            if text is None:
                text = spec.storage_format_text
        if kwargs_names is None:
            kwargs_names = self.signature_keyword_names
        if args_values is None:
            args_values = tuple(
                self._get(_)
                for _ in self.signature_positional_names
                )
        if args_values:
            kwargs_names = list(kwargs_names)
            names = self.signature_positional_names
            if not self.signature_accepts_args:
                names += self.signature_keyword_names
            names = names[:len(args_values)]
            for name in names:
                if name in kwargs_names:
                    kwargs_names.remove(name)
            kwargs_names = tuple(kwargs_names)
        return dict(
            args_values=args_values,
            as_storage_format=as_storage_format,
            is_bracketed=is_bracketed,
            is_indented=is_indented,
            kwargs_names=kwargs_names,
            text=text,
            via=via,
            )

    @staticmethod
    def _get_module_path_parts(subject):
        if isinstance(subject, type):
            class_ = subject
        elif type(subject) is subject.__class__:
            class_ = type(subject)
        class_name = class_.__name__
        parts = class_.__module__.split('.')
        while parts and parts[-1] == class_name:
            parts.pop()
        parts.append(class_name)
        return parts

    @staticmethod
    def _get_types(subject, result=None):
        """
        Gets all non-builtin types referenced in storage format.

        ..  container:: example

            >>> flute = abjad.Flute()

            >>> types = abjad.StorageFormatManager._get_types(flute)
            >>> for type_ in types:
            ...     type_
            ...
            <class 'abjad.instruments.Flute'>
            <class 'abjad.markups.Markup'>
            <class 'abjad.pitch.Accidental.Accidental'>
            <class 'abjad.pitch.NamedPitch.NamedPitch'>
            <class 'abjad.pitch.Octave.Octave'>
            <class 'abjad.pitch.PitchRange.PitchRange'>

        ..  container:: example

            >>> dictionary = abjad.OrderedDict(
            ...     item_class=abjad.NamedPitch,
            ...     )

            >>> types = abjad.StorageFormatManager._get_types(dictionary)
            >>> for _ in types:
            ...     _
            ...
            <class 'abjad.pitch.NamedPitch.NamedPitch'>
            <class 'abjad.utilities.OrderedDict.OrderedDict'>

        Returns tuple of types.
        """
        import abjad
        type_type = type
        if result is None:
            result = set()
        agent = StorageFormatManager(subject)
        if isinstance(subject, str):
            return []
        arguments = []
        if not isinstance(subject, type_type):
            if hasattr(subject, '_get_storage_format_specification'):
                specification = subject._get_storage_format_specification()
                keyword_argument_names = specification.keyword_argument_names
                if keyword_argument_names is None:
                    keyword_argument_names = agent.signature_keyword_names
                for name in keyword_argument_names:
                    value = getattr(subject, name)
                    arguments.append(value)
                positional_argument_values = specification.positional_argument_values
                if positional_argument_values is None:
                    signature = agent.inspect_signature(subject)
                    names, _, _, _ = signature
                    positional_argument_values = [
                        agent._get(name) for name in names]
                arguments.extend(positional_argument_values)
            else:
                arguments.extend(agent.get_template_dict().values())
        if isinstance(subject, collections.Mapping):
            for key, value in subject.items():
                result.update(agent._get_types(key))
                result.update(agent._get_types(value))
        elif isinstance(subject, collections.Iterable):
            for value in subject:
                result.update(agent._get_types(value))
        arguments.append(subject)
        for argument in arguments:
            if not isinstance(argument, type_type):
                if argument is not subject:
                    result.update(agent._get_types(argument))
            if isinstance(argument, type_type):
                if not issubclass(argument, abjad.AbjadObject):
                    continue
            elif type(argument).__module__ in (
                'builtins',
                '__builtin__',
                'abc',
                ):
                continue
            if not isinstance(argument, type_type):
                if argument is not subject:
                    result.update(agent._get_types(argument))
                argument = type(argument)
            result.add(argument)
        result = sorted(result, key=lambda x: (x.__module__, x.__name__))
        return result

    def _get_whitespace(self, is_indented):
        if is_indented:
            return self._indented_whitespace
        return self._unindented_whitespace

    @staticmethod
    def _make_hashable(value):
        if isinstance(value, dict):
            value = tuple(value.items())
        elif isinstance(value, list):
            value = tuple(value)
        elif isinstance(value, (set, frozenset)):
            value = tuple(value)
        return value

    def _map_positional_values_to_names(self, values):
        names = self.signature_positional_names
        if not self.signature_accepts_args:
            names += self.signature_keyword_names
        names = names[:len(values)]
        return names

    ### PUBLIC PROPERTIES ###

    @property
    def client(self):
        return self._client

    @property
    def format_specification(self):
        from abjad import system
        if self._format_specification is None:
            if (
                not isinstance(self._client, type) and
                hasattr(self._client, '_get_format_specification')
                ):
                self._format_specification = \
                    self._client._get_format_specification()
            else:
                self._format_specification = system.FormatSpecification(
                    self._client)
        return self._format_specification

    @property
    def signature_accepts_args(self):
        return self._signature_accepts_args

    @property
    def signature_accepts_kwargs(self):
        return self._signature_accepts_kwargs

    @property
    def signature_keyword_names(self):
        return self._signature_keyword_names

    @property
    def signature_names(self):
        return (
            self.signature_positional_names +
            self.signature_keyword_names
            )

    @property
    def signature_positional_names(self):
        return self._signature_positional_names

    ### PUBLIC METHODS ###

    def get_class_name_prefix(
        self,
        as_storage_format,
        include_root_package=None,
        ):
        agent = StorageFormatManager(self._client)
        if not isinstance(self._client, type):
            class_name = type(self._client).__name__
        else:
            class_name = self._client.__name__
        if as_storage_format:
            root_package_name = self.get_root_package_name()
            root_package = importlib.import_module(root_package_name)
            parts = [root_package_name]
            if class_name not in dir(root_package):
                tools_package_name = agent.get_tools_package_name()
                parts.append(tools_package_name)
            parts.append(class_name)
            return '.'.join(parts)
        return class_name

    def get_hash_values(self):
        values = []
        if isinstance(self._client, type):
            values.append(self._client)
        else:
            values.append(type(self._client))
        template_items = sorted(self.get_template_dict().items())
        values.extend(self._make_hashable(v) for k, v in template_items)
        return tuple(values)

    def get_import_statements(self):
        """
        Gets import statements.

        ..  container:: example

            >>> flute = abjad.Flute()
            >>> abjad.f(flute)
            abjad.Flute(
                name='flute',
                short_name='fl.',
                markup=abjad.Markup(
                    contents=['Flute'],
                    ),
                short_markup=abjad.Markup(
                    contents=['Fl.'],
                    ),
                allowable_clefs=('treble',),
                context='Staff',
                middle_c_sounding_pitch=abjad.NamedPitch("c'"),
                pitch_range=abjad.PitchRange('[C4, D7]'),
                primary=True,
                )

            >>> agent = abjad.StorageFormatManager(flute)
            >>> for line in agent.get_import_statements():
            ...     line
            ...
            'from abjad import instruments'
            'from abjad import markups'
            'from abjad import pitch'

        Returns tuple of strings.
        """
        import_statements = set()
        classes = self._get_types(self.client)
        for class_ in classes:
            agent = StorageFormatManager(class_)
            root_package_name = agent.get_root_package_name()
            if root_package_name in ('builtins', '__builtin__'):
                continue
            elif root_package_name == 'abjad':
                tools_package_name = agent.get_tools_package_name()
                import_statement = 'from abjad import {}'.format(
                    tools_package_name)
            else:
                import_statement = 'import {}'.format(root_package_name)
            import_statements.add(import_statement)
        return tuple(sorted(import_statements))

    def get_repr_format(self):
        assert hasattr(self._client, '_get_format_specification')
        pieces = self._format_specced_object(
            as_storage_format=False,
            )
        return ''.join(pieces)

    def get_repr_keyword_dict(self):
        from abjad import system
        names = self.specification.repr_kwargs_names
        if names is None:
            specification = system.StorageFormatSpecification(self.client)
            names = specification.keyword_argument_names or ()
        keyword_dict = {}
        for name in names:
            keyword_dict[name] = self._get(name)
        return keyword_dict

    def get_repr_positional_values(self):
        from abjad import system
        values = self.specification.repr_args_values
        if values is None:
            specification = system.StorageFormatSpecification(self.client)
            values = specification.positional_argument_values or ()
        return tuple(values)

    def get_root_package_name(self):
        return self._get_module_path_parts(self._client)[0]

    def get_storage_format(self):
        assert (
            '_get_storage_format_specification' in dir(self._client) or
            hasattr(self._client, '_get_storage_format_specification') or
            hasattr(self._client, '_get_format_specification')
            )
        pieces = self._format_specced_object(
            as_storage_format=True,
            )
        result = ''.join(pieces)
        return result

    def get_storage_format_keyword_dict(self):
        from abjad import system
        names = self.specification.storage_format_kwargs_names
        if names is None:
            if hasattr(self.client, '_get_storage_format_specification'):
                specification = self.client._get_storage_format_specification()
            else:
                specification = system.StorageFormatSpecification(self.client)
            names = specification.keyword_argument_names or ()
        keyword_dict = {}
        for name in names:
            keyword_dict[name] = self._get(name)
        return keyword_dict

    def get_storage_format_positional_values(self):
        from abjad import system
        values = self.specification.storage_format_args_values
        if values is None:
            if hasattr(self.client, '_get_storage_format_specification'):
                specification = self.client._get_storage_format_specification()
            else:
                specification = system.StorageFormatSpecification(self.client)
            values = specification.positional_argument_values or ()
        return tuple(values)

    def get_template_dict(self):
        template_names = self.format_specification.template_names
        if template_names is None:
            template_names = self.signature_names
            # TODO: This will be factored out when SFS/SFM are removed.
            if hasattr(self.client, '_get_storage_format_specification'):
                specification = self.client._get_storage_format_specification()
                template_names.extend(
                    specification._keyword_argument_names or ()
                    )
            else:
                template_names.extend(
                    self.format_specification.storage_format_kwargs_names or ())
            template_names = sorted(set(template_names))
        template_dict = collections.OrderedDict()
        for name in template_names:
            template_dict[name] = self._get(name)
        return template_dict

    def get_tools_package_name(self):
        parts = self._get_module_path_parts(self._client)
        if parts[0] in ('abjad', 'abjadext', 'ide'):
            for part in reversed(parts):
                if part == parts[-1]:
                    continue
                return part
        return '.'.join(parts[:-1])

    @classmethod
    def inspect_signature(class_, subject):
        positional_names = []
        keyword_names = []
        accepts_args = False
        accepts_kwargs = False
        if not isinstance(subject, type):
            subject = type(subject)
        try:
            signature = inspect.signature(subject)
        except ValueError:
            return (
                positional_names,
                keyword_names,
                accepts_args,
                accepts_kwargs,
                )
        for name, parameter in signature.parameters.items():
            if parameter.kind == inspect._POSITIONAL_OR_KEYWORD:
                if parameter.default == parameter.empty:
                    positional_names.append(name)
                else:
                    keyword_names.append(name)
            # Python 3 allow keyword only parameters:
            elif (
                hasattr(inspect, '_KEYWORD_ONLY') and
                parameter.kind == inspect._KEYWORD_ONLY
                ):
                keyword_names.append(name)
            elif parameter.kind == inspect._VAR_POSITIONAL:
                accepts_args = True
            elif parameter.kind == inspect._VAR_KEYWORD:
                accepts_kwargs = True
        return (
            positional_names,
            keyword_names,
            accepts_args,
            accepts_kwargs,
            )



import collections
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.utilities.CyclicTuple import CyclicTuple


class Duplication(AbjadValueObject):
    """
    Duplication.

    ..  container:: example:

        >>> operator_ = abjad.Duplication(counts=2, period=4)

        >>> abjad.f(operator_)
        abjad.Duplication(
            counts=2,
            period=4,
            )

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_counts',
        '_indices',
        '_period',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        counts=None,
        indices=None,
        period=None,
        ):
        if counts is not None:
            if isinstance(counts, collections.Sequence):
                assert len(counts)
                counts = tuple(int(_) for _ in counts)
                assert all(0 <= _ for _ in counts)
            else:
                counts = int(counts)
                assert 0 <= counts
        self._counts = counts
        if indices is not None:
            assert all(isinstance(_, int) for _ in indices), repr(indices)
            indices = tuple(indices)
        self._indices = indices
        if period is not None:
            period = int(period)
            assert 0 < period
        self._period = period

    ### SPECIAL METHODS ###

    def __call__(self, argument):
        """
        Calls rotation on `argument`.

        ..  container:: example

            Duplicates once without period:

            >>> operator_ = abjad.Duplication(counts=1)
            >>> numbers = [1, 2, 3, 4]
            >>> operator_(numbers)
            [1, 2, 3, 4, 1, 2, 3, 4]

        ..  container:: example

            Duplicates twice without period:

            >>> operator_ = abjad.Duplication(counts=2)
            >>> pitch_classes = abjad.PitchClassSegment([0, 1, 4, 7])
            >>> operator_(pitch_classes)
            PitchClassSegment([0, 1, 4, 7, 0, 1, 4, 7, 0, 1, 4, 7])

        ..  container:: example

            Duplicates periodically:

            >>> operator_ = abjad.Duplication(counts=1, period=3)
            >>> pitches = abjad.PitchSegment("c' d' e' f' g' a' b' c''")
            >>> for pitch in operator_(pitches):
            ...     pitch
            ...
            NamedPitch("c'")
            NamedPitch("d'")
            NamedPitch("e'")
            NamedPitch("c'")
            NamedPitch("d'")
            NamedPitch("e'")
            NamedPitch("f'")
            NamedPitch("g'")
            NamedPitch("a'")
            NamedPitch("f'")
            NamedPitch("g'")
            NamedPitch("a'")
            NamedPitch("b'")
            NamedPitch("c''")
            NamedPitch("b'")
            NamedPitch("c''")

        ..  container:: example

            Duplicate indices:

            >>> operator_ = abjad.Duplication(
            ...     counts=1,
            ...     indices=(0, -1),
            ...     )
            >>> pitch_classes = abjad.PitchClassSegment([0, 1, 4, 7])
            >>> operator_(pitch_classes)
            PitchClassSegment([0, 0, 1, 4, 7, 7])

        ..  container:: example

            Duplicate indices periodically:

            >>> operator_ = abjad.Duplication(
            ...     counts=1,
            ...     indices=(0,),
            ...     period=2,
            ...     )
            >>> pitch_classes = abjad.PitchClassSegment([0, 1, 4, 7, 9])
            >>> operator_(pitch_classes)
            PitchClassSegment([0, 0, 1, 4, 4, 7, 9, 9])

        ..  container:: example

            Duplicate indices periodically with different counts:

            >>> operator_ = abjad.Duplication(
            ...     counts=(1, 2),
            ...     indices=(0,),
            ...     period=2,
            ...     )
            >>> pitch_classes = abjad.PitchClassSegment([0, 1, 4, 7, 9])
            >>> operator_(pitch_classes)
            PitchClassSegment([0, 0, 1, 4, 4, 4, 7, 9, 9])

        ..  container:: example

            Cyclic counts:

            >>> operator_ = abjad.Duplication(counts=(0, 1, 2, 3))
            >>> pitch_classes = abjad.PitchClassSegment([0, 1, 4, 7, 9])
            >>> operator_(pitch_classes)
            PitchClassSegment([0, 1, 1, 4, 4, 4, 7, 7, 7, 7, 9])

        Returns new object with type equal to that of `argument`.
        """
        import abjad

        if not isinstance(argument, collections.Sequence):
            argument = (argument,)

        counts = self.counts
        if isinstance(counts, int):
            counts = counts + 1
        else:
            counts = [_ + 1 for _ in counts]

        if not self.period and not self.indices:
            if isinstance(counts, int):
                return type(argument)(argument * counts)
            else:
                counts = CyclicTuple(counts)
                result = []
                for i, x in enumerate(argument):
                    count = counts[i]
                    result.extend([x] * count)
                if isinstance(argument, abjad.TypedCollection):
                    result = abjad.new(argument, items=result)
                else:
                    result = type(argument)(result)
                return result

        if isinstance(counts, int):
            counts = [counts]
        counts = CyclicTuple(counts)

        if not self.indices:
            if isinstance(argument, abjad.TypedCollection):
                result = abjad.new(argument, items=())
            else:
                result = type(argument)()
            iterator = abjad.sequence(argument).partition_by_counts(
                [self.period],
                cyclic=True,
                overhang=True,
                )
            for i, shard in enumerate(iterator):
                shard = type(argument)(shard) * counts[i]
                result = result + shard
            return result

        pattern = abjad.Pattern(
            indices=self.indices,
            period=self.period,
            )
        result = []
        length = len(argument)
        j = 0
        for i, x in enumerate(argument):
            if pattern.matches_index(i, length):
                count = counts[j]
                result.extend([x] * count)
                j += 1
            else:
                result.append(x)
        if isinstance(argument, abjad.TypedCollection):
            result = abjad.new(argument, items=result)
        else:
            result = type(argument)(result)
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def counts(self):
        """
        Gets counts of duplication.

        ..  container:: example

            >>> operator_ = abjad.Duplication(counts=1, period=3)
            >>> operator_.counts
            1

        Returns integer or none.
        """
        return self._counts

    @property
    def indices(self):
        """
        Gets indices of duplication.

        ..  container:: example

            >>> operator_ = abjad.Duplication(
            ...     counts=1,
            ...     indices=(0, -1),
            ...     )
            >>> operator_.indices
            (0, -1)

        Returns integer or none.
        """
        return self._indices

    @property
    def period(self):
        """
        Gets period of duplication.

        ..  container:: example

            >>> operator_ = abjad.Duplication(counts=1, period=3)
            >>> operator_.period
            3

        Returns integer or none.
        """
        return self._period



"""
Tools for modeling spanners: slurs, beams, glissandi, etc.
"""

from .Spanner import Spanner
from .Beam import Beam
from .BowContactSpanner import BowContactSpanner
from .Glissando import Glissando
from .Hairpin import Hairpin
from .Hairpin import hairpin
from .HorizontalBracket import HorizontalBracket
from .OctavationSpanner import OctavationSpanner
from .PhrasingSlur import PhrasingSlur
from .PianoPedalSpanner import PianoPedalSpanner
from .Slur import Slur
from .TextSpanner import TextSpanner
from .TextSpanner import text_spanner
from .Tie import Tie
from .TrillSpanner import TrillSpanner



def all_are_nonnegative_integer_powers_of_two(argument):
    """
    Is true when ``argument`` is an iterable collection of nonnegative
    integer powers of two.

    ..  container:: example

        >>> items = [0, 1, 1, 1, 2, 4, 32, 32]
        >>> abjad.mathtools.all_are_nonnegative_integer_powers_of_two(items)
        True

        >>> abjad.mathtools.all_are_nonnegative_integer_powers_of_two(17)
        False

    ..  container:: example

        Is true when ``argument`` is empty:

        >>> abjad.mathtools.all_are_nonnegative_integer_powers_of_two([])
        True

    Returns true or false.
    """
    from abjad import mathtools
    try:
        return all(
            mathtools.is_nonnegative_integer_power_of_two(_)
            for _ in argument
            )
    except TypeError:
        return False



from abjad.system.AbjadObject import AbjadObject


class BenchmarkScoreMaker(AbjadObject):
    """
    Benchmark score-maker.

    ..  container:: example

        >>> benchmark_score_maker = abjad.BenchmarkScoreMaker()

        >>> benchmark_score_maker
        BenchmarkScoreMaker()

    Use to instantiate scores for benchmark testing.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Benchmarking'

    __slots__ = ()

    ### PUBLIC METHODS ###

    def make_bound_hairpin_score_01(self):
        """
        Make 200-note voice with p-to-f bound crescendo spanner
        on every 4 notes.

        2.12 (r9726) initialization:        279,448 function calls

        2.12 (r9726) LilyPond format:       124,517 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.Sequenc(voice[:]).partition_by_counts(
            [4],
            cyclic=True,
            ):
            crescendo = abjad.Hairpin('<')
            abjad.attach(crescendo, part)
            dynamic = abjad.Dynamic('p')
            abjad.attach(dynamic, part[0])
            dynamic = abjad.Dynamic('r')
            abjad.attach(dynamic, part[-1])
        return voice

    def make_bound_hairpin_score_02(self):
        """
        Make 200-note voice with p-to-f bound crescendo spanner
        on every 20 notes.

        2.12 (r9726) initialization:        268,845 function calls

        2.12 (r9726) LilyPond format:       117,846 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [20],
            cyclic=True,
            ):
            crescendo = abjad.Hairpin('<')
            abjad.attach(crescendo, part)
            dynamic = abjad.Dynamic('p')
            abjad.attach(dynamic, part[0])
            dynamic = abjad.Dynamic('r')
            abjad.attach(dynamic, part[-1])
        return voice

    def make_bound_hairpin_score_03(self):
        """
        Make 200-note voice with p-to-f bound crescendo spanner
        on every 100 notes.

        2.12 (r9726) initialization:        267,417 function calls

        2.12 (r9726) LilyPond format:       116,534 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [100],
            cyclic=True,
            ):
            crescendo = abjad.Hairpin('<')
            abjad.attach(crescendo, part)
            dynamic = abjad.Dynamic('p')
            abjad.attach(dynamic, part[0])
            dynamic = abjad.Dynamic('r')
            abjad.attach(dynamic, part[-1])
        return voice

    def make_hairpin_score_01(self):
        """
        Make 200-note voice with crescendo spanner on every 4 notes.

        2.12 (r9726) initialization:        248,502 function calls
        2.12 (r9728) initialization:        248,502 function calls

        2.12 (r9726) LilyPond format:       138,313 function calls
        2.12 (r9728) LilyPond format:       134,563 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [4],
            cyclic=True,
            ):
            crescendo = abjad.Hairpin('<')
            abjad.attach(crescendo, part)
        return voice

    def make_hairpin_score_02(self):
        """
        Make 200-note voice with crescendo spanner on every 20 notes.

        2.12 (r9726) initialization:        248,687 function calls
        2.12 (r9728) initialization:        248,687 function calls

        2.12 (r9726) LilyPond format:       134,586 function calls
        2.12 (r9728) LilyPond format:       129,836 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [20],
            cyclic=True,
            ):
            crescendo = abjad.Hairpin('<')
            abjad.attach(crescendo, part)
        return voice

    def make_hairpin_score_03(self):
        """
        Make 200-note voice with crescendo spanner on every 100 notes.

        2.12 (r9726) initialization:        249,363 function calls
        2.12 (r9726) initialization:        249,363 function calls

        2.12 (r9726) LilyPond format:       133,898 function calls
        2.12 (r9728) LilyPond format:       128,948 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [100],
            cyclic=True,
            ):
            crescendo = abjad.Hairpin('<')
            abjad.attach(crescendo, part)
        return voice

    def make_score_00(self):
        """
        Make 200-note voice (with nothing else).

        2.12 (r9710) initialization:        156,821 function calls
        2.12 (r9726) initialization:        156,827 function calls

        2.12 (r9703) LilyPond format:        99,127 function calls
        2.12 (r9710) LilyPond format:       100,126 function calls
        2.12 (r9726) LilyPond format:       105,778 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        return voice

    def make_score_with_indicators_01(self):
        """
        Make 200-note voice with dynamic on every 20th note:

        2.12 (r9704) initialization:        630,433 function calls
        2.12 (r9710) initialization:        235,120 function calls
        2.12 r(9726) initialization:        235,126 function calls

        2.12 (r9704) LilyPond format:       136,637 function calls
        2.12 (r9710) LilyPond format:        82,730 function calls
        2.12 (r9726) LilyPond format:        88,382 function calls

        """
        import abjad
        staff = abjad.Staff(200 * abjad.Note("c'16"))
        for part in abjad.sequence(staff[:]).partition_by_counts(
            [20],
            cyclic=True,
            ):
            dynamic = abjad.Dynamic('f')
            abjad.attach(dynamic, part[0])
        return staff

    def make_score_with_indicators_02(self):
        """
        Make 200-note staff with dynamic on every 4th note.

        2.12 (r9704) initialization:      4,632,761 function calls
        2.12 (r9710) initialization:        327,280 function calls
        2.12 (r9726) initialization:        325,371 function calls

        2.12 (r9704) LilyPond format:       220,277 function calls
        2.12 (r9710) LilyPond format:        84,530 function calls
        2.12 (r9726) LilyPond format:        90,056 function calls

        """
        import abjad
        staff = abjad.Staff(200 * abjad.Note("c'16"))
        for part in abjad.sequence(staff[:]).partition_by_counts(
            [4],
            cyclic=True,
            ):
            dynamic = abjad.Dynamic('f')
            abjad.attach(dynamic, part[0])
        return staff

    def make_score_with_indicators_03(self):
        """
        Make 200-note staff with dynamic on every note.

        2.12 (r9704) initialization:     53,450,195 function calls (!!)
        2.12 (r9710) initialization:      2,124,500 function calls
        2.12 (r9724) initialization:      2,122,591 function calls

        2.12 (r9704) LilyPond format:       533,927 function calls
        2.12 (r9710) LilyPond format:        91,280 function calls
        2.12 (r9724) LilyPond format:        96,806 function calls

        """
        import abjad
        staff = abjad.Staff(200 * abjad.Note("c'16"))
        selector = abjad.select().leaves()
        for note in selector(staff):
            dynamic = abjad.Dynamic('f')
            abjad.attach(dynamic, note)
        return staff

    def make_spanner_score_01(self):
        """
        Make 200-note voice with durated complex beam spanner on every 4 notes.

        2.12 (r9710) initialization:        248,654 function calls
        2.12 (r9724) initialization:        248,660 function calls

        2.12 (r9703) LilyPond format:       425,848 function calls
        2.12 (r9710) LilyPond format:       426,652 function calls
        2.12 (r9724) LilyPond format:       441,884 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [4],
            cyclic=True,
            ):
            beam = abjad.Beam()
            abjad.attach(beam, part)
        return voice

    def make_spanner_score_02(self):
        """
        Make 200-note voice with durated complex beam spanner
        on every 20 notes.

        2.12 (r9710) initialization:        250,954 function calls
        2.12 (r9724) initialization:        248,717 function calls

        2.12 (r9703) LilyPond format:       495,768 function calls
        2.12 (r9710) LilyPond format:       496,572 function calls
        2.12 (r9724) LilyPond format:       511,471 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [20],
            cyclic=True,
            ):
            beam = abjad.Beam()
            abjad.attach(beam, part)
        return voice

    def make_spanner_score_03(self):
        """
        Make 200-note voice with durated complex beam spanner
        on every 100 notes.

        2.12 (r9710) initialization:        251,606 function calls
        2.12 (r9724) initialization:        249,369 function calls

        2.12 (r9703) LilyPond format:       509,752 function calls
        2.12 (r9710) LilyPond format:       510,556 function calls
        2.12 (r9724) LilyPond format:       525,463 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [100],
            cyclic=True,
            ):
            beam = abjad.Beam()
            abjad.attach(beam, part)
        return voice

    def make_spanner_score_04(self):
        """
        Make 200-note voice with slur spanner on every 4 notes.

        2.12 (r9724) initialization:        245,683 function calls

        2.12 (r9703) LilyPond format:       125,577 function calls
        2.12 (r9724) LilyPond format:       111,341 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [4],
            cyclic=True,
            ):
            slur = abjad.Slur()
            abjad.attach(slur, part)
        return voice

    def make_spanner_score_05(self):
        """
        Make 200-note voice with slur spanner on every 20 notes.

        2.12 (r9724) initialization:        248,567 function calls

        2.12 (r9703) LilyPond format:       122,177 function calls
        2.12 (r9724) LilyPond format:       107,486 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [20],
            cyclic=True,
            ):
            slur = abjad.Slur()
            abjad.attach(slur, part)
        return voice

    def make_spanner_score_06(self):
        """
        Make 200-note voice with slur spanner on every 100 notes.

        2.12 (r9724) initialization:        249,339 function calls

        2.12 (r9703) LilyPond format:       121,497 function calls
        2.12 (r9724) LilyPond format:       106,718 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [100],
            cyclic=True,
            ):
            slur = abjad.Slur()
            abjad.attach(slur, part)
        return voice

    def make_spanner_score_07(self):
        """
        Make 200-note voice with (vanilla) beam spanner on every 4 notes.

        2.12 (r9724) initialization:        245,683 function calls

        2.12 (r9703) LilyPond format:       125,577 function calls
        2.12 (r9724) LilyPond format:       132,556 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [4],
            cyclic=True,
            ):
            beam = abjad.Beam()
            abjad.attach(beam, part)
        return voice

    def make_spanner_score_08(self):
        """
        Make 200-note voice with (vanilla) beam spanner on every 20 notes.

        2.12 (r9724) initialization:        248,567 function calls

        2.12 (r9703) LilyPond format:       122,177 function calls
        2.12 (r9724) LilyPond format:       129,166 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [20],
            cyclic=True,
            ):
            beam = abjad.Beam()
            abjad.attach(beam, part)
        return voice

    def make_spanner_score_09(self):
        """
        Make 200-note voice with (vanilla) beam spanner on every 100 notes.

        2.12 (r9724) initialization:        249,339 function calls

        2.12 (r9703) LilyPond format:       121,497 function calls
        2.12 (r9724) LilyPond format:       128,494 function calls

        """
        import abjad
        voice = abjad.Voice(200 * abjad.Note("c'16"))
        for part in abjad.sequence(voice[:]).partition_by_counts(
            [100],
            cyclic=True,
            ):
            beam = abjad.Beam()
            abjad.attach(beam, part)
        return voice



from abjad import mathtools
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.FormatSpecification import FormatSpecification
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle
from abjad.system.StorageFormatManager import StorageFormatManager


class StemTremolo(AbjadValueObject):
    """
    Stem tremolo.

    ..  container:: example

        Sixteenth-note tremolo:

        >>> note = abjad.Note("c'4")
        >>> stem_tremolo = abjad.StemTremolo(16)
        >>> abjad.attach(stem_tremolo, note)
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(note)
            c'4
            :16

    ..  container:: example

        Thirty-second-note tremolo:

        >>> note = abjad.Note("c'4")
        >>> stem_tremolo = abjad.StemTremolo(32)
        >>> abjad.attach(stem_tremolo, note)
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(note)
            c'4
            :32

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_tremolo_flags',
        )

    _format_slot = 'after'

    ### INITIALIZER ###

    def __init__(self, tremolo_flags: int = 16) -> None:
        if isinstance(tremolo_flags, type(self)):
            tremolo_flags = tremolo_flags.tremolo_flags
        tremolo_flags = int(tremolo_flags)
        if not mathtools.is_nonnegative_integer_power_of_two(tremolo_flags):
            message = 'nonnegative integer power of 2: {tremolo_flags!r}.'
            raise ValueError(message)
        self._tremolo_flags = tremolo_flags

    ### SPECIAL METHODS ###

    def __format__(self, format_specification='') -> str:
        """
        Formats stem tremolo.

        ..  container:: example

            Sixteenth-note tremolo:

            >>> stem_tremolo = abjad.StemTremolo(16)
            >>> print(format(stem_tremolo))
            :16

        ..  container:: example

            Thirty-second-note tremolo:

            >>> stem_tremolo = abjad.StemTremolo(32)
            >>> print(format(stem_tremolo))
            :32

        """
        if format_specification in ('', 'lilypond'):
            return self._get_lilypond_format()
        assert format_specification == 'storage'
        return StorageFormatManager(self).get_storage_format()

    def __str__(self) -> str:
        """
        Gets string representation of stem tremolo.

        ..  container:: example

            Sixteenth-note tremolo:

            >>> stem_tremolo = abjad.StemTremolo(16)
            >>> print(str(stem_tremolo))
            :16

        ..  container:: example

            Thirty-second-note tremolo:

            >>> stem_tremolo = abjad.StemTremolo(32)
            >>> print(str(stem_tremolo))
            :32

        """
        return f':{self.tremolo_flags!s}'

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return FormatSpecification(
            client=self,
            storage_format_is_indented=False,
            )

    def _get_lilypond_format(self):
        return str(self)

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        bundle.after.stem_tremolos.append(self._get_lilypond_format())
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def tremolo_flags(self) -> int:
        """
        Gets tremolo flags of stem tremolo.

        ..  container:: example

            Sixteenth-note tremolo:

            >>> stem_tremolo = abjad.StemTremolo(16)
            >>> stem_tremolo.tremolo_flags
            16

        ..  container:: example

            Thirty-second-note tremolo:

            >>> stem_tremolo = abjad.StemTremolo(32)
            >>> stem_tremolo.tremolo_flags
            32

        Set to nonnegative integer power of 2.
        """
        return self._tremolo_flags

    @property
    def tweaks(self) -> None:
        r"""
        Are not implemented on stem tremolo.
        
        The LilyPond ``:`` command refuses tweaks.

        Override the LilyPond ``StemTremolo`` grob instead.
        """
        pass



from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle


class Repeat(AbjadValueObject):
    r"""
    Repeat.

    ..  container:: example

        Volta repeat:

        >>> container = abjad.Container("c'4 d'4 e'4 f'4")
        >>> repeat = abjad.Repeat()
        >>> abjad.attach(repeat, container)
        >>> staff = abjad.Staff([container])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score)  # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    \repeat volta 2
                    {
                        c'4
                        d'4
                        e'4
                        f'4
                    }
                }
            >>

    ..  container:: example

        Unfold repeat:

        >>> container = abjad.Container("c'4 d'4 e'4 f'4")
        >>> repeat = abjad.Repeat(repeat_type='unfold')
        >>> abjad.attach(repeat, container)
        >>> staff = abjad.Staff([container])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score)  # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    \repeat unfold 2
                    {
                        c'4
                        d'4
                        e'4
                        f'4
                    }
                }
            >>

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_repeat_count',
        '_repeat_type',
        )

    _can_attach_to_containers = True

    _context = 'Score'

    _format_leaf_children = False

    _format_slot = 'before'

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        repeat_count: int = 2,
        repeat_type: str = 'volta',
        ) -> None:
        repeat_count = int(repeat_count)
        assert 1 < repeat_count
        self._repeat_count = repeat_count
        assert repeat_type in ('volta', 'unfold')
        self._repeat_type = repeat_type

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        r"""
        Gets string representation of repeat.

        ..  container:: example

            Volta repeat:

            >>> str(abjad.Repeat())
            '\\repeat volta 2'

        ..  container:: example

            Unfold repeat:

            >>> str(abjad.Repeat(repeat_type='unfold'))
            '\\repeat unfold 2'

        """
        return rf'\repeat {self.repeat_type} {self.repeat_count}'

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        return str(self)

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        bundle.before.commands.append(self._get_lilypond_format())
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def context(self):
        """
        Gets (historically conventional) context.

        ..  container:: example

            Volta repeat:

            >>> repeat = abjad.Repeat()
            >>> repeat.context
            'Score'

        ..  container:: example

            Unfold repeat:

            >>> repeat = abjad.Repeat(repeat_type='unfold')
            >>> repeat.context
            'Score'

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def repeat_count(self) -> int:
        """
        Gets repeat count of repeat.

        ..  container:: example

            Volta repeat:

            >>> repeat = abjad.Repeat()
            >>> repeat.repeat_count
            2

        ..  container:: example

            Unfold repeat:

            >>> repeat = abjad.Repeat(repeat_type='unfold')
            >>> repeat.repeat_count
            2

        """
        return self._repeat_count

    @property
    def repeat_type(self) -> str:
        """
        Gets repeat type of repeat.

        ..  container:: example

            Volta repeat:

            >>> repeat = abjad.Repeat()
            >>> repeat.repeat_type
            'volta'

        ..  container:: example

            Unfold repeat:

            >>> repeat = abjad.Repeat(repeat_type='unfold')
            >>> repeat.repeat_type
            'unfold'

        """
        return self._repeat_type

    @property
    def tweaks(self) -> None:
        r"""
        Are not implemented on repeat.
        
        The LilyPond ``\repeat`` command refuses tweaks.

        Override the LilyPond ``BarLine`` grob instead.
        """
        pass



from abjad.utilities.Enumerator import Enumerator
from .Vector import Vector


class IntervalVector(Vector):
    """
    Interval vector.

    ..  container:: example

        Initializes from pitch segment:

        >>> pitch_segment = abjad.PitchSegment(
        ...     items=[0, 11, 7, 4, 2, 9, 3, 8, 10, 1, 5, 6],
        ...     )
        >>> numbered_interval_vector = abjad.IntervalVector(
        ...     items=pitch_segment,
        ...     item_class=abjad.NumberedInterval,
        ...     )
        >>> for interval, count in sorted(numbered_interval_vector.items(),
        ...     key=lambda x: (x[0].direction_number, x[0].number)):
        ...     print(interval, count)
        ...
        -11 1
        -10 1
        -9 1
        -8 2
        -7 3
        -6 3
        -5 4
        -4 4
        -3 4
        -2 5
        -1 6
        +1 5
        +2 5
        +3 5
        +4 4
        +5 3
        +6 3
        +7 2
        +8 2
        +9 2
        +10 1

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        import abjad
        if isinstance(items, (
            abjad.PitchSegment,
            abjad.PitchSet,
            abjad.PitchClassSegment,
            abjad.PitchClassSet,
            )):
            intervals = []
            enumerator = Enumerator(items)
            pairs = enumerator.yield_pairs()
            for first, second in pairs:
                intervals.append(second - first)
            items = intervals
        Vector.__init__(
            self,
            items=items,
            item_class=item_class,
            )

    ### SPECIAL METHODS ###

    def __repr__(self):
        """
        Gets interpret representation of interval vector.

        ..  container:: example

            Gets interpreter representation of interval vector:

            >>> pitch_segment = abjad.PitchSegment(
            ...     items=[0, 11, 7, 4, 2, 9, 3, 8, 10, 1, 5, 6],
            ...     )
            >>> vector = abjad.IntervalVector(
            ...     items=pitch_segment,
            ...     item_class=abjad.NumberedInterval,
            ...     )

            >>> vector
            IntervalVector({-11: 1, -10: 1, -9: 1, -8: 2, -7: 3, -6: 3, -5: 4, -4: 4, -3: 4, -2: 5, -1: 6, 1: 5, 2: 5, 3: 5, 4: 4, 5: 3, 6: 3, 7: 2, 8: 2, 9: 2, 10: 1}, item_class=NumberedInterval)

        ..  container:: example

            Initializes from interpreter representation of interval vector:

            >>> abjad.IntervalVector(vector)
            IntervalVector({-11: 1, -10: 1, -9: 1, -8: 2, -7: 3, -6: 3, -5: 4, -4: 4, -3: 4, -2: 5, -1: 6, 1: 5, 2: 5, 3: 5, 4: 4, 5: 3, 6: 3, 7: 2, 8: 2, 9: 2, 10: 1}, item_class=NumberedInterval)

        Returns string.
        """
        return super().__repr__()

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        import abjad
        return abjad.NamedInterval

    @property
    def _numbered_item_class(self):
        import abjad
        return abjad.NumberedInterval

    @property
    def _parent_item_class(self):
        import abjad
        return abjad.Interval

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(
        class_,
        selection,
        item_class=None,
        ):
        """
        Makes interval vector from `selection`.

        Returns interval vector.
        """
        import abjad
        pitch_segment = abjad.PitchSegment.from_selection(selection)
        return class_(
            pitch_segment,
            item_class=item_class,
            )



import copy
from .Context import Context


class Score(Context):
    r"""
    Score.

    ..  container:: example

        >>> staff_1 = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> staff_2 = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> score = abjad.Score([staff_1, staff_2])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    c'8
                    d'8
                    e'8
                    f'8
                }
                \new Staff
                {
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Contexts'

    __slots__ = ()

    _default_lilypond_type = 'Score'

    ### INITIALIZER ###

    def __init__(
        self,
        components=None,
        lilypond_type='Score',
        is_simultaneous=True,
        name=None,
        ):
        Context.__init__(
            self,
            components=components,
            lilypond_type=lilypond_type,
            is_simultaneous=is_simultaneous,
            name=name,
            )

    ### PUBLIC METHODS ###

    def add_final_bar_line(
        self,
        abbreviation='|.',
        to_each_voice=False,
        ):
        r"""
        Add final bar line to end of score.


            >>> staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> score = abjad.Score([staff])
            >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }
            >>


        >>> bar_line = score.add_final_bar_line()
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    c'4
                    d'4
                    e'4
                    f'4
                    \bar "|." %! SCORE1
                }
            >>

        Set ``to_each_voice`` to true to make part extraction easier.

        Returns bar line.
        """
        import abjad
        bar_line = abjad.BarLine(abbreviation)
        if not to_each_voice:
            last_leaf = abjad.inspect(self).leaf(-1)
            abjad.attach(bar_line, last_leaf, tag='SCORE1')
        else:
            for voice in abjad.iterate(self).components(abjad.Voice):
                last_leaf = abjad.inspect(voice).leaf(-1)
                abjad.attach(bar_line, last_leaf, tag='SCORE1')
        return bar_line

    def add_final_markup(self, markup, extra_offset=None):
        r"""
        Adds ``markup`` to end of score.

        ..  container:: example

            Adds markup to last leaf:

            >>> staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> score = abjad.Score([staff])
            >>> place = abjad.Markup('Bremen - Boston - LA.', direction=abjad.Down)
            >>> date = abjad.Markup('July 2010 - May 2011.')
            >>> markup = abjad.Markup.right_column([place, date], direction=abjad.Down)
            >>> markup = markup.italic()
            >>> markup = score.add_final_markup(
            ...     markup,
            ...     extra_offset=(0.5, -2),
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(score)
                \new Score
                <<
                    \new Staff
                    {
                        c'4
                        d'4
                        e'4
                        \once \override TextScript.extra-offset = #'(0.5 . -2)
                        f'4
                        _ \markup { %! SCORE2
                            \italic %! SCORE2
                                \right-column %! SCORE2
                                    { %! SCORE2
                                        "Bremen - Boston - LA." %! SCORE2
                                        "July 2010 - May 2011." %! SCORE2
                                    } %! SCORE2
                            } %! SCORE2
                    }
                >>

        ..  container:: example

            Adds markup to last multimeasure rest:

            >>> staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> staff.append(abjad.MultimeasureRest((4, 4)))
            >>> score = abjad.Score([staff])
            >>> place = abjad.Markup(
            ...     'Bremen - Boston - LA.',
            ...     direction=abjad.Down,
            ...     )
            >>> date = abjad.Markup('July 2010 - May 2011.')
            >>> markup = abjad.Markup.right_column(
            ...     [place, date],
            ...     direction=abjad.Down,
            ...     )
            >>> markup = markup.italic()
            >>> markup = score.add_final_markup(
            ...     markup,
            ...     extra_offset=(14.5, -2),
            ...     )
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(score)
                \new Score
                <<
                    \new Staff
                    {
                        c'4
                        d'4
                        e'4
                        f'4
                        \once \override MultiMeasureRestText.extra-offset = #'(14.5 . -2)
                        R1
                        _ \markup { %! SCORE2
                            \italic %! SCORE2
                                \right-column %! SCORE2
                                    { %! SCORE2
                                        "Bremen - Boston - LA." %! SCORE2
                                        "July 2010 - May 2011." %! SCORE2
                                    } %! SCORE2
                            } %! SCORE2
                    }
                >>

        Returns none.
        """
        import abjad
        selection = abjad.select(self)
        last_leaf = selection._get_component(abjad.Leaf, -1)
        markup = copy.copy(markup)
        abjad.attach(markup, last_leaf, tag='SCORE2')
        if extra_offset is not None:
            if isinstance(last_leaf, abjad.MultimeasureRest):
                grob_proxy = abjad.override(last_leaf).multi_measure_rest_text
            else:
                grob_proxy = abjad.override(last_leaf).text_script
            grob_proxy.extra_offset = extra_offset
        return markup

    @staticmethod
    def make_piano_score(leaves=None, lowest_treble_pitch='B3', sketch=False):
        r"""
        Makes piano score from ``leaves``.

        ..  container:: example

            Makes empty piano score:

            >>> result = abjad.Score.make_piano_score()

            >>> abjad.f(result[0])
            \new Score
            <<
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                    }
                    \context Staff = "Bass Staff"
                    {
                    }
                >>
            >>

        ..  container:: example

            Makes piano score from leaves:

            >>> notes = [abjad.Note(x, (1, 4)) for x in [-12, 37, -10, 2, 4, 17]]
            >>> result = abjad.Score.make_piano_score(leaves=notes)
            >>> abjad.show(result[0]) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(result[0])
                \new Score
                <<
                    \new PianoStaff
                    <<
                        \context Staff = "Treble Staff"
                        {
                            \clef "treble"
                            r4
                            cs''''4
                            r4
                            d'4
                            e'4
                            f''4
                        }
                        \context Staff = "Bass Staff"
                        {
                            \clef "bass"
                            c4
                            r4
                            d4
                            r4
                            r4
                            r4
                        }
                    >>
                >>

        ..  container:: example

            Makes piano sketch score from leaves:

            >>> maker = abjad.NoteMaker()
            >>> notes = maker(
            ...     [-12, -10, -8, -7, -5, 0, 2, 4, 5, 7],
            ...     [(1, 16)],
            ...     )
            >>> result = abjad.Score.make_piano_score(
            ...     leaves=notes,
            ...     sketch=True,
            ...     )
            >>> abjad.show(result[0]) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(result[0])
                \new Score
                \with
                {
                    \override BarLine.stencil = ##f
                    \override BarNumber.transparent = ##t
                    \override SpanBar.stencil = ##f
                    \override TimeSignature.stencil = ##f
                }
                <<
                    \new PianoStaff
                    <<
                        \context Staff = "Treble Staff"
                        {
                            \clef "treble"
                            r16
                            r16
                            r16
                            r16
                            r16
                            c'16
                            d'16
                            e'16
                            f'16
                            g'16
                        }
                        \context Staff = "Bass Staff"
                        {
                            \clef "bass"
                            c16
                            d16
                            e16
                            f16
                            g16
                            r16
                            r16
                            r16
                            r16
                            r16
                        }
                    >>
                >>


        Returns score, treble staff, bass staff triple.
        """
        import abjad
        leaves = leaves or []
        lowest_treble_pitch = abjad.NamedPitch(lowest_treble_pitch)
        treble_staff = abjad.Staff(name='Treble Staff')
        bass_staff = abjad.Staff(name='Bass Staff')
        staff_group = abjad.StaffGroup(
            [treble_staff, bass_staff],
            lilypond_type='PianoStaff',
            )
        score = abjad.Score()
        score.append(staff_group)
        for leaf in leaves:
            treble_pitches, bass_pitches = [], []
            for pitch in abjad.inspect(leaf).pitches():
                if pitch < lowest_treble_pitch:
                    bass_pitches.append(pitch)
                else:
                    treble_pitches.append(pitch)
            written_duration = leaf.written_duration
            if not treble_pitches:
                treble_leaf = abjad.Rest(written_duration)
            elif len(treble_pitches) == 1:
                treble_leaf = abjad.Note(treble_pitches[0], written_duration)
            else:
                treble_leaf = abjad.Chord(treble_pitches, written_duration)
            treble_staff.append(treble_leaf)
            if not bass_pitches:
                bass_leaf = abjad.Rest(written_duration)
            elif len(bass_pitches) == 1:
                bass_leaf = abjad.Note(bass_pitches[0], written_duration)
            else:
                bass_leaf = abjad.Chord(bass_pitches, written_duration)
            bass_staff.append(bass_leaf)
        if 0 < len(treble_staff):
            abjad.attach(abjad.Clef('treble'), treble_staff[0])
        if 0 < len(bass_staff):
            abjad.attach(abjad.Clef('bass'), bass_staff[0])
        if sketch:
            abjad.override(score).time_signature.stencil = False
            abjad.override(score).bar_number.transparent = True
            abjad.override(score).bar_line.stencil = False
            abjad.override(score).span_bar.stencil = False
        return score, treble_staff, bass_staff



"""
Command-line tools.
"""

import fnmatch
import importlib
import inspect
import os
import pathlib
import re
import shutil
import types
from abjad import utilities
from uqbar.cli import CLI, CLIAggregator


class AbjDevScript(CLIAggregator):
    '''
    Entry-point to the Abjad developer scripts catalog.

    Can be accessed on the commandline via `abj-dev` or `ajv`:

    ..  shell::

        ajv --help

    `ajv` supports subcommands similar to `svn`:

    ..  shell::

        ajv api --help

    '''

    ### CLASS VARIABLES ###

    config_name = '.abjadrc'
    short_description = 'Entry-point to Abjad developer scripts catalog.'

    ### SPECIAL METHODS ###

    @property
    def cli_classes(self):
        """
        Lists CLI classes for aggregation.
        """
        def scan_module(module):
            classes = []
            for name in dir(module):
                obj = getattr(module, name)
                if not isinstance(obj, type):
                    continue
                elif not issubclass(obj, CLI):
                    continue
                elif issubclass(obj, type(self)):
                    continue
                elif inspect.isabstract(obj):
                    continue
                classes.append(obj)
            return classes
        import abjad.cli
        classes = scan_module(abjad.cli)
        try:
            import abjadext  # type: ignore
            for abjadext_path in abjadext.__path__:
                abjadext_path = pathlib.Path(abjadext_path)
                for extension_path in abjadext_path.iterdir():
                    if (
                        extension_path.name.startswith(('.', '_')) or
                        not extension_path.is_dir() or
                        not (extension_path / '__init__.py').exists()
                    ):
                        continue
                    module_name = 'abjadext.{}'.format(extension_path.name)
                    module = importlib.import_module(module_name)
                    classes.extend(scan_module(module))
        except ImportError:
            pass
        classes.sort(key=lambda x: x.__name__)
        return classes


class CleanScript(CLI):
    '''
    Removes *.pyc, *.swp files and __pycache__ and tmp* directories recursively
    in a path.

    ..  shell::

        ajv clean --help

    '''

    ### CLASS VARIABLES ###

    alias = 'clean'
    config_name = '.abjadrc'
    short_description = (
        'Clean *.pyc, *.swp, .cache,  __pycache__ and tmp* '
        'files and folders from PATH.'
        )

    ### PRIVATE METHODS ###

    def _process_args(self, arguments):
        if (
            not arguments.cache and
            not arguments.pyc and
            not arguments.pycache and
            not arguments.swp and
            not arguments.tmp
        ):
            arguments.cache = True,
            arguments.pyc = True
            arguments.pycache = True
            arguments.swp = True
            arguments.tmp = True
        print('Cleaning...')
        if arguments.pyc:
            print('\t*.pyc files')
        if arguments.swp:
            print('\t*.swp files')
        if arguments.cache:
            print('\t.cache directories')
        if arguments.pycache:
            print('\t__pycache__ directories')
        if arguments.tmp:
            print('\ttmp* directories')

        for root_directory, directory_names, file_names in os.walk(
            arguments.path):
            if '.svn' in directory_names:
                directory_names.remove('.svn')
            extensions = ()
            if arguments.pyc:
                extensions += ('.pyc',)
            if arguments.swp:
                extensions += ('.swp',)
            for file_name in file_names:
                if file_name.endswith(extensions):
                    file_path = os.path.join(
                        root_directory,
                        file_name,
                        )
                    os.remove(file_path)
            directories_to_remove = []
            for directory_name in directory_names:
                directory = os.path.join(
                    root_directory,
                    directory_name,
                    )
                should_remove = False
                if arguments.cache:
                    if directory_name == '.cache':
                        should_remove = True
                if arguments.pycache:
                    if directory_name == '__pycache__':
                        should_remove = True
                if arguments.tmp:
                    if directory_name.startswith('tmp'):
                        should_remove = True
                if not os.listdir(directory):
                    should_remove = True
                if should_remove:
                    shutil.rmtree(directory)
                    directories_to_remove.append(directory_name)
            for directory_name in directories_to_remove:
                directory_names.remove(directory_name)

    def _setup_argument_parser(self, parser):
        parser.add_argument(
            'path',
            default=os.getcwd(),
            help='directory tree to be recursed over',
            nargs='?',
            type=self._validate_path,
            )
        parser.add_argument(
            '--cache',
            action='store_true',
            help='delete *.cache folders',
            )
        parser.add_argument(
            '--pyc',
            action='store_true',
            help='delete *.pyc files',
            )
        parser.add_argument(
            '--pycache',
            action='store_true',
            help='delete __pycache__ folders',
            )
        parser.add_argument(
            '--swp',
            action='store_true',
            help='delete Vim *.swp file',
            )
        parser.add_argument(
            '--tmp',
            action='store_true',
            help='delete tmp* folders',
            )


class ReplaceScript(CLI):
    '''
    Replaces text in files recursively.

    ..  shell::

        ajv replace text --help

    Multiple patterns for excluding files or folders can be specified by
    restating the `--without-files` or `--without-dirs` commands:

    ..  code-block:: bash

        abjad$ ajv replace text . foo bar -F *.txt -F *.rst -F *.htm

    '''

    ### CLASS VARIABLES ###

    alias = 'replace'
    config_name = '.abjadrc'
    short_description = 'Replace text.'

    ### PRIVATE METHODS ###

    def _get_naive_search_callable(self, arguments):

        class NaiveSearch(object):

            def __init__(self, pattern):
                self.pattern = pattern

            def __call__(self, line, pos):
                index = line.find(self.pattern, pos)
                if 0 <= index:
                    return index, len(self.pattern)
                return -1, 0

        return NaiveSearch(arguments.old)

    def _get_regex_search_callable(self, arguments):

        class RegexSearch(object):

            def __init__(self, pattern, escaped=False, whole_words_only=False):
                try:
                    if escaped:
                        pattern = re.escape(pattern)
                    if whole_words_only:
                        pattern += r'\b'
                    self.pattern = re.compile(pattern)
                    self.whole_words_only = whole_words_only
                except Exception:
                    message = "can't compile {!r} as a regex pattern."
                    message = message.format(pattern)
                    raise ValueError(message)

            def __call__(self, line, pos):
                start, length = self._search(line, pos)
                if self.whole_words_only and 0 < start:
                    while start != -1 and (
                        line[start - 1].isalnum() or line[start - 1] == '_'):
                        start, length = self._search(line, start + length)
                return start, length

            def _search(self, line, pos):
                match = self.pattern.search(line, pos)
                if match is None:
                    return -1, 0
                return match.start(), match.end() - match.start()

        return RegexSearch(
            arguments.old,
            escaped=not arguments.regex,
            whole_words_only=arguments.whole_words_only,
            )

    def _process_args(self, arguments):
        import abjad
        message = 'Replacing {!r} with {!r} ...'
        message = message.format(arguments.old, arguments.new)
        print(message)
        skipped_dirs_patterns = self.skipped_directories
        skipped_dirs_patterns += arguments.without_dirs
        skipped_files_patterns = self.skipped_files + arguments.without_files
        if (arguments.regex or
            (not arguments.regex and arguments.whole_words_only)
            ):
            arguments.old = self._get_regex_search_callable(arguments)
            index, length = arguments.old('', 0)
            if 0 <= index:
                message = 'regex pattern {!r} matches the empty string.'
                message = message.format(arguments.old.pattern.pattern)
                raise ValueError(message)
        else:
            arguments.old = self._get_naive_search_callable(arguments)
        changed_file_count = 0
        changed_line_count = 0
        changed_item_count = 0
        for root, dirs, files in os.walk(arguments.path):
            dirs_to_remove = []
            for dir in dirs:
                for pattern in skipped_dirs_patterns:
                    if fnmatch.fnmatch(dir, pattern):
                        dirs_to_remove.append(dir)
                        break
            for dir in dirs_to_remove:
                dirs.remove(dir)
            for file in sorted(files):
                valid = True
                for pattern in skipped_files_patterns:
                    if fnmatch.fnmatch(file, pattern):
                        valid = False
                        break
                if not valid:
                    continue
                changed_lines, changed_items = self._process_file(
                    arguments, os.path.join(root, file))
                if changed_lines:
                    changed_file_count += 1
                    changed_line_count += changed_lines
                    changed_item_count += changed_items
        print()
        item_identifier = abjad.String('instance').pluralize(changed_item_count)
        line_identifier = abjad.String('line').pluralize(changed_line_count)
        file_identifier = abjad.String('file').pluralize(changed_file_count)
        message = '\tReplaced {} {} over {} {} in {} {}.'
        message = message.format(
            changed_item_count,
            item_identifier,
            changed_line_count,
            line_identifier,
            changed_file_count,
            file_identifier,
            )
        print(message)

    def _process_file(self, arguments, path):
        changed_items = 0
        changed_lines = 0
        try:
            with open(path, 'r') as f:
                lines = f.read().split('\n')
        except UnicodeDecodeError:
            return changed_lines, changed_items
        results = []
        for i, line in enumerate(lines):
            line, changes = self._process_line(
                line, i, path, arguments.old, arguments.new, arguments.force, arguments.verbose)
            results.append(line)
            if changes:
                changed_items += changes
                changed_lines += 1
        if results != lines:
            with open(path, 'w') as f:
                f.write('\n'.join(results))
        return changed_lines, changed_items

    def _process_line(
        self,
        line,
        line_number,
        file_name,
        search,
        replacement,
        force,
        verbose,
        ):
        index, changes = 0, 0
        index, length = search(line, index)
        while 0 <= index:
            should_replace = False
            replaced_line = line[:index] + replacement + line[index + length:]
            carats = (' ' * index) + ('^' * length)
            if force:
                should_replace = True
                if verbose:
                    print()
                    print('{}: {}'.format(file_name, line_number))
                    print('-{}'.format(line))
                    print('+{}'.format(replaced_line))
            else:
                print()
                print('{}: {}'.format(file_name, line_number))
                print()
                print('{}'.format(line))
                print('{}'.format(carats))
                print()
                result = input('Replace? [Y/n] > ').lower()
                while result not in ('', 'y', 'yes', 'n', 'no'):
                    result = input('Replace? [Y/n] > ').lower()
                if result in ('', 'y', 'yes'):
                    should_replace = True
            if should_replace:
                index += length - (length - len(replacement))
                line = replaced_line
                changes += 1
            else:
                index += length
            index, length = search(line, index)
        return line, changes

    def _setup_argument_parser(self, parser):
        parser.add_argument(
            'path',
            default=os.getcwd(),
            help='directory tree to be recursed over',
            nargs='?',
            type=self._validate_path,
            )
        parser.add_argument(
            'old',
            help='old text',
            )
        parser.add_argument(
            'new',
            help='new text',
            )
        parser.add_argument(
            '--verbose',
            action='store_true',
            help='print replacement info even when --force flag is set.',
            )
        parser.add_argument(
            '-Y', '--force',
            action='store_true',
            help='force "yes" to every replacement'
            )
        parser.add_argument(
            '-R', '--regex',
            action='store_true',
            help='treat "old" as a regular expression',
            )
        parser.add_argument(
            '-W', '--whole-words-only',
            action='store_true',
            help='''match only whole words, similar to grep's "-w" flag''',
            )
        parser.add_argument(
            '-F', '--without-files',
            action='append',
            default=[],
            help='Exclude files matching pattern(s)',
            metavar='PATTERN',
            )
        parser.add_argument(
            '-D', '--without-dirs',
            action='append',
            default=[],
            help='Exclude folders matching pattern(s)',
            metavar='PATTERN',
            )

    ### PUBLIC PROPERTIES ###

    @property
    def skipped_directories(self):
        r'''Skipped directories.

        Returns list.
        '''
        return [
            '.svn',
            '.git',
            'build'
            ]

    @property
    def skipped_files(self):
        r'''Skipped files.

        Returns list.
        '''
        return [
            __file__,
            self.program_name,
            '*.ai',
            '*.backup',
            '*.doc',
            '*.docx',
            '*.doctree',
            '*.gif',
            '*.indd',
            '*.indt',
            '*.jpg',
            '*.jpeg',
            '*.mid',
            '*.midi',
            '*.nb',
            '*.pages',
            '*.pdf',
            '*.pickle',
            '*.pkl',
            '*.png',
            '*.ps',
            '*.psd',
            '*.pyc',
            '*.rtf',
            '*.tif',
            '*.tiff',
            '*.txt',
            '*.wav',
            '*.zip',
            '.DS_Store',
            ]


class StatsScript(CLI):
    '''
    Builds statistics about a codebase.

    ..  shell::

        ajv stats  --help

    '''

    ### CLASS VARIABLES ###

    alias = 'stats'
    config_name = '.abjadrc'
    short_description = 'Build statistics about Python modules in PATH.'

    ### PRIVATE METHODS ###

    def _iterate_module(self, module):
        results = []
        for name in dir(module):
            obj = getattr(module, name)
            if not isinstance(obj, (type, types.FunctionType)):
                continue
            elif obj.__module__ != module.__name__:
                continue
            results.append(obj)
        return results

    def _print_results(self, counts):
        template = utilities.String.normalize('''
            Source lines: {source_lines}
            Public classes: {public_classes}
                Unique public methods: {unique_public_methods}
                Unique public properties: {unique_public_properties}
                Unique private methods: {unique_private_methods}
                Unique private properties: {unique_private_properties}
            Public functions: {public_functions}
            Private classes: {private_classes}
            Private functions: {private_functions}
        ''')
        result = template.format(
            source_lines=counts['source_lines'],
            public_classes=counts['public_classes'],
            unique_public_methods=counts['unique_public_methods'],
            unique_public_properties=counts['unique_public_properties'],
            unique_private_methods=counts['unique_private_methods'],
            unique_private_properties=counts['unique_private_properties'],
            public_functions=counts['public_functions'],
            private_classes=counts['private_classes'],
            private_functions=counts['private_functions'],
            )
        print(result)

    def _process_args(self, arguments):
        from abjad import utilities
        path = arguments.path
        if not os.path.isdir(path):
            path = os.path.dirname(path)
        counts = self._setup_counts()
        for module in utilities.yield_all_modules(
            code_root=path,
            ignored_file_names=[],
            ):
            with open(module.__file__, 'r') as file_pointer:
                contents = file_pointer.read()
                counts['source_lines'] += contents.count('\n')
            for obj in self._iterate_module(module):
                if isinstance(obj, types.FunctionType):
                    if obj.__name__.startswith('_'):
                        counts['private_functions'] += 1
                    else:
                        counts['public_functions'] += 1
                elif isinstance(obj, type):
                    if obj.__name__.startswith('_'):
                        counts['private_classes'] += 1
                    else:
                        counts['public_classes'] += 1
                        for attr in inspect.classify_class_attrs(obj):
                            if attr.defining_class != obj:
                                continue
                            if attr.kind in ('method', 'class method', 'static method'):
                                if attr.name.startswith('_'):
                                    counts['unique_private_methods'] += 1
                                else:
                                    counts['unique_public_methods'] += 1
                            elif attr.kind in ('property,'):
                                if attr.name.startswith('_'):
                                    counts['unique_private_properties'] += 1
                                else:
                                    counts['unique_public_properties'] += 1

        self._print_results(counts)

    def _setup_argument_parser(self, parser):
        parser.add_argument(
            'path',
            default=os.getcwd(),
            help='directory tree to be recursed over',
            nargs='?',
            type=self._validate_path,
            )

    def _setup_counts(self):
        counts = {
            'source_lines': 0,
            'private_classes': 0,
            'private_functions': 0,
            'public_classes': 0,
            'public_functions': 0,
            'unique_public_methods': 0,
            'unique_public_properties': 0,
            'unique_private_methods': 0,
            'unique_private_properties': 0,
            }
        return counts


def run_ajv():
    r'''Entry point for setuptools.

    One-line wrapper around AbjDevScript.
    '''
    AbjDevScript()()



import copy
from ply import lex  # type: ignore
from abjad import core
from abjad import exceptions
from abjad.system.AbjadObject import AbjadObject


# TODO: should not inherit from AbjadObject because no slots
class LilyPondLexicalDefinition(AbjadObject):
    """
    The lexical definition of LilyPond's syntax.

    Effectively equivalent to LilyPond's ``lexer.ll`` file.

    Not composer-safe.

    Used internally by ``LilyPondParser``.
    """

    ### INITIALIZER ###

    def __init__(self, client=None):
        self.client = client

    states = (
        # lexer.ll:115
        # ('extratoken', 'exclusive'),
        # ('chords', 'exclusive'),
        # ('figures', 'exclusive'),
        # ('incl', 'exclusive'),
        # ('lyrics', 'exclusive'),
        # ('lyric_quote ', 'exclusive'),
        ('longcomment', 'exclusive'),
        ('markup', 'exclusive'),
        ('notes', 'exclusive'),
        ('quote', 'exclusive'),
        # ('sourcefileline', 'exclusive'),
        # ('sourcefilename', 'exclusive'),
        ('version', 'exclusive'),
        ('scheme', 'exclusive'),
    )

    # lexer.ll:129
    A               = r'[a-zA-Z\200-\377]'
    AA              = r'(%s|_)' % A
    N               = r'[0-9]'
    AN              = r'(%s|%s)' % (AA, N)
    ANY_CHAR        = r'(.|\n)'
    PUNCT           = r"[?!:'`]"
    ACCENT          = r'''\\[`'"^]'''
    NATIONAL        = r'[\001-\006\021-\027\031\036]'
    TEX             = r'%s|-|%s|%s|%s' % (AA, PUNCT, ACCENT, NATIONAL)
    WORD            = r'%s%s*' % (A, AN)
    DASHED_WORD     = r'%s(%s|-)*' % (A, AN)
    DASHED_KEY_WORD = r'\\%s' % DASHED_WORD

    # lexer.ll:144
    ALPHAWORD       = r'%s+' % A
    DIGIT           = r'%s' % N
    UNSIGNED        = r'%s+' % N
    INT             = r'(-?%s)' % UNSIGNED
    REAL            = r'((%s\.%s*)|(-?\.%s+))' % (INT, N, N)
    E_UNSIGNED      = r'\\%s+' % N
    FRACTION        = r'%s+\/%s+' % (N, N)
    KEYWORD         = r'\\%s' % WORD
    WHITE           = r'[ \n\t\f\r]' # only whitespace
    HORIZONTALWHITE = r'[ \t]' # only non-line-breaking whitespace
    BLACK           = r'[^ \n\t\f\r]' # only non-whitespace
    RESTNAME        = r'[rs]'
    NOTECOMMAND     = r'\\%s+' % A
    MARKUPCOMMAND   = r'\\(%s|[-_])+' % A
    LYRICS          = r'(%s|%s)[^0-9 \t\n\r\f]*' % (AA, TEX)
    ESCAPED         = r'''[nt\\'"]'''
    EXTENDER        = r'__'
    HYPHEN          = r'--'
    BOM_UTF8        = r'\357\273\277'

    keywords = {
        # parser.yy:182, lily-lexer.cc:39
#        '\\accepts': 'ACCEPTS',
#        '\\addlyrics': 'ADDLYRICS',
#        '\\alias': 'ALIAS',
#        '\\alternative': 'ALTERNATIVE',
#        '\\book': 'BOOK',
#        '\\bookpart': 'BOOKPART',
        '\\change': 'CHANGE',
#        '\\chordmode': 'CHORDMODE',
#        '\\chords': 'CHORDS',
#        '\\consists': 'CONSISTS',
        '\\context': 'CONTEXT',
        '\\default': 'DEFAULT',
#        '\\defaultchild': 'DEFAULTCHILD',
#        '\\denies': 'DENIES',
#        '\\description': 'DESCRIPTION',
#        '\\drummode': 'DRUMMODE',
#        '\\drums': 'DRUMS',
#        '\\figuremode': 'FIGUREMODE',
#        '\\figures': 'FIGURES',
        '\\header': 'HEADER',
#        '\\version-error': 'INVALID',
        '\\layout': 'LAYOUT',
#        '\\lyricmode': 'LYRICMODE',
#        '\\lyrics': 'LYRICS',
#        '\\lyricsto': 'LYRICSTO',
        '\\markup': 'MARKUP',
        '\\markuplist': 'MARKUPLIST',
        '\\midi': 'MIDI',
#        '\\name': 'NAME',
#        '\\notemode': 'NOTEMODE',
        '\\override': 'OVERRIDE',
        '\\paper': 'PAPER',
#        '\\remove': 'REMOVE',
#        '\\repeat': 'REPEAT',
        '\\rest': 'REST',
        '\\revert': 'REVERT',
        '\\score': 'SCORE',
        '\\sequential': 'SEQUENTIAL',
        '\\set': 'SET',
        '\\simultaneous': 'SIMULTANEOUS',
        '\\tempo': 'TEMPO',
#        '\\type': 'TYPE',
        '\\unset': 'UNSET',
        '\\with': 'WITH',

        # parser.yy:233
        '\\new': 'NEWCONTEXT',

        # ???
#        '\\objectid': 'OBJECTID',
    }

    tokens = [
#        'CHORD_BASS', # "/+"
#        'CHORD_CARET', # "^"
#        'CHORD_COLON', # ":"
#        'CHORD_MINUS', # "-"
#        'CHORD_SLASH', # "/"
        'ANGLE_OPEN', # "<"
        'ANGLE_CLOSE', # ">"
        'DOUBLE_ANGLE_OPEN', # "<<"
        'DOUBLE_ANGLE_CLOSE', # ">>"
        'E_BACKSLASH', # "\\"
        'E_ANGLE_CLOSE', # "\\>"
#        'E_CHAR', # "\\C[haracter]"
        'E_CLOSE', # "\\)"
        'E_EXCLAMATION', # "\\!"
#        'E_BRACKET_OPEN', # "\\["
        'E_OPEN', # "\\("
#        'E_BRACKET_CLOSE', # "\\]"
        'E_ANGLE_OPEN', # "\\<"
#        'E_PLUS', # "\\+"
#        'E_TILDE', # "\\~"
        'EXTENDER', # "__"

#        'FIGURE_CLOSE', # "\\>"
#        'FIGURE_OPEN', # "\\<"
#        'FIGURE_SPACE', # "_"
        'HYPHEN', # "--"

#        'LYRIC_MARKUP',
        'MULTI_MEASURE_REST',

        'E_UNSIGNED',
        'UNSIGNED',

        'EXPECT_MARKUP', # "markup?"
        'EXPECT_PITCH', # "ly:pitch?"
        'EXPECT_DURATION', # "ly:duration?"
        'EXPECT_SCM', # "scheme?"
        'BACKUP', # "(backed-up?)"
        'REPARSE', # "(reparsed?)"
        'EXPECT_MARKUP_LIST', # "markup-list?"
        'EXPECT_OPTIONAL', # "optional?"
        'EXPECT_NO_MORE_ARGS', #

#        'EMBEDDED_LILY', # "#{"

#        'BOOK_IDENTIFIER',
#        'CHORD_MODIFIER',
        'CHORD_REPETITION',
        'CONTEXT_DEF_IDENTIFIER',
        'CONTEXT_MOD_IDENTIFIER',
#        'DRUM_PITCH',
        'PITCH_IDENTIFIER',
        'DURATION_IDENTIFIER',
        'EVENT_IDENTIFIER',
        'EVENT_FUNCTION',
        'FRACTION',
#        'LYRICS_STRING',
#        'LYRIC_ELEMENT',
#        'LYRIC_MARKUP_IDENTIFIER',
        'MARKUP_FUNCTION',
        'MARKUP_LIST_FUNCTION',
        'MARKUP_IDENTIFIER',
        'MARKUPLIST_IDENTIFIER',
        'MUSIC_FUNCTION',
        'MUSIC_IDENTIFIER',
        'NOTENAME_PITCH',
        'NUMBER_IDENTIFIER',
        'OUTPUT_DEF_IDENTIFIER',
        'REAL',
        'RESTNAME',
        'SCM_FUNCTION',
        'SCM_IDENTIFIER',
        'SCM_TOKEN',
        'SCORE_IDENTIFIER',
        'STRING',
        'STRING_IDENTIFIER',
        'TONICNAME_PITCH',
    ] + list(keywords.values())

    literals = (
        '!', "'", '(', ')', '*', '+', ',', '-',
        '.', '/', ':', '<', '=', '>', '?', '[',
        '\\', '^', '_', '{', '|', '}', '~', ']',
    )

    string_accumulator = ''

    ### LEXICAL RULES ###

    # lexer.ll:165
    # <*>\r
    def t_ANY_165(self, t):
        r'\r'
        pass

    # lexer.ll:169
    # <extratoken>{ANY_CHAR}

    # lexer.ll:186
    # <extratoken><<EOF>>

    # lexer.ll:201
    # <INITIAL,chords,lyrics,figures,notes>{BOM_UTF8}/.*

    # lexer.ll:210
    # <INITIAL,chords,figures,incl,lyrics,markup,notes>"%{"
    def t_INITIAL_markup_notes_210(self, t):
        r'%{'
        t.lexer.push_state('longcomment')
        pass

    # lexer.ll:214
    # <INITIAL,chords,figures,incl,lyrics,markup,notes>%[^{\n\r][^\n\r]*[\n\r]
    def t_INITIAL_markup_notes_214(self, t):
        r'%[^{\n\r][^\n\r]*[\n\r]'
        pass

    def t_INITIAL_markup_notes_214_EOF(self, t):
        r'%[^{\n\r][^\n\r]*$'
        pass

    #lexer.ll:216
    # <INITIAL,chords,figures,incl,lyrics,markup,notes>%[^{\n\r]
    def t_INITIAL_markup_notes_216(self, t):
        r'%[^{\n\r]'
        pass

    #lexer.ll:218
    # <INITIAL,chords,figures,incl,lyrics,markup,notes>%[\n\r]
    def t_INITIAL_markup_notes_218(self, t):
        r'%[\n\r]'
        pass

    # lexer.ll:220
    # <INITIAL,chords,figures,incl,lyrics,markup,notes>%[^{\n\r][^\n\r]*
    def t_INITIAL_markup_notes_220(self, t):
        r'%[^{\n\r][^\n\r]*'
        pass

    # lexer.ll:222
    # <INITIAL,chords,figures,incl,lyrics,markup,notes>{WHITE}+
    def t_INITIAL_markup_notes_222(self, t):
        '[ \n\t\f\r]'
        pass

    # lexer.ll:227
    # <INITIAL,notes,figures,chords,markup>\"
    def t_INITIAL_markup_notes_227(self, t):
        r'\"'
        t.lexer.push_state('quote')
        self.string_accumulator = ''
        pass

    # lexer.ll:233
    # <INITIAL,chords,lyrics,notes,figures>\\version{WHITE}*
    def t_INITIAL_notes_233(self, t):
        r'\\version[ \n\t\f\r]*'
        t.lexer.push_state('version')
        pass

    # lexer.ll:236
    # <INITIAL,chords,lyrics,notes,figures>\\sourcefilename{WHITE}*

    # lexer.ll:239
    # <INITIAL,chords,lyrics,notes,figures>\\sourcefileline{WHITE}*

    # lexer.ll:242
    # <version>\"[^"]*\"
    def t_version_242(self, t):
        r'\"[^"]*\"'
        t.lexer.pop_state()
        pass

    # lexer.ll:256
    # <sourcefilename>\"[^"]*\"

    # lexer.ll:270
    # <sourcefileline>{INT}

    # lexer.ll:278
    # <version>{ANY_CHAR}
    @lex.TOKEN(ANY_CHAR)
    def t_version_278(self, t):
        print(("LilyPondParser: Illegal character '%s'" % t.value[0]))
        t.lexer.skip(1)

    # lexer.ll:282
    # <sourcefilename>{ANY_CHAR}

    # lexer.ll:286
    # <sourcefileline>{ANY_CHAR}

    # lexer.ll:296
    # <longcomment>"%"+"}"
    def t_longcomment_296(self, t):
        r'%}'
        t.lexer.pop_state()
        pass

    # lexer.ll:291
    # <longcomment>[^\%]*
    def t_longcomment_291(self, t):
        r'[^%]+'
        pass

    # lexer.ll:293
    # <longcomment>\%*[^}%]*
    def t_longcomment_293(self, t):
        r'%+[^}%]*'
        pass

    # lexer.ll:302
    # <INITIAL,chords,lyrics,notes,figures>\\maininput

    # lexer.ll:312
    # <INITIAL,chords,lyrics,figures,notes>\\include

    # lexer.ll:315
    # <incl>\"[^"]*\"

    # lexer.ll:322
    # <incl>\\{BLACK}*{WHITE}?

    # lexer.ll:341
    # <incl,version,sourcefilename>\"[^"]*
    def t_version_341(self, t):
        r'"[^"]*'
        message = 'end quote missing: {!r}.'
        message = message.format(t)
        raise Exception(message)

    # lexer.ll:345
    # <chords,notes,figures>{RESTNAME}
#    @lex.token.TOKEN(RESTNAME)
#    def t_notes_345(self, t):
#        t.type = 'RESTNAME'
#        return t

    # lexer.ll:350
    # <chords,notes,figures>R
#    def t_notes_350(self, t):
#        'R'
#        t.type = 'MULTI_MEASURE_REST'
#        return t

    #def t_INITIAL_markup_notes_353_boolean(self, t):
    #    '\#\#(t|f)'
    #    t.type = 'SCM_TOKEN'
    #    if t.value[2] == 't':
    #        t.value = True
    #    else:
    #        t.value = False
    #    return t

    #@lex.TOKEN("\#'%s" % DASHED_WORD)
    #def t_INITIAL_markup_notes_353_identifier(self, t):
    #    t.type = 'SCM_IDENTIFIER'
    #    t.value = t.value[2:]
    #    return t

    # lexer.ll:353
    # <INITIAL,chords,figures,lyrics,markup,notes>#
    def t_INITIAL_markup_notes_353(self, t):
        '\#'
        from abjad import parser as abjad_parser
        #t.type = 'SCHEME_START'
        #t.lexer.push_state('INITIAL')
        scheme_parser = abjad_parser.SchemeParser(debug=False)
        input_string = t.lexer.lexdata[t.lexpos+1:]
        #print 'PREPARSE'
        try:
            scheme_parser(input_string)
        except exceptions.SchemeParserFinishedError:
            result = scheme_parser.result
            cursor_end = scheme_parser.cursor_end
            #print 'PARSED: {!r}'.format(input_string[:cursor_end])
            t.value = result
            t.type = 'SCM_TOKEN'
            #if isinstance(result, str):
            #    t.type = 'STRING'
            #    if t.value.find(' ') != -1:
            #        t.value = '"{}"'.format(t.value)
            #else:
            #    t.type = 'SCM_TOKEN'
            t.lexer.skip(scheme_parser.cursor_end + 1)
        return t

    # lexer.ll:387
    # <INITIAL,notes,lyrics>\<\<
    def t_INITIAL_notes_387(self, t):
        r'\<\<'
        t.type = 'DOUBLE_ANGLE_OPEN'
        return t

    # lexer.ll:390
    # <INITIAL,notes,lyrics>\>\>
    def t_INITIAL_notes_390(self, t):
        r'\>\>'
        t.type = 'DOUBLE_ANGLE_CLOSE'
        return t

    # lexer.ll:396
    # <INITIAL,notes>\<
    def t_INITIAL_notes_396(self, t):
        r'\<'
        t.type = 'ANGLE_OPEN'
        return t

    # lexer.ll:399
    # <INITIAL,notes>\>
    def t_INITIAL_notes_399(self, t):
        r'\>'
        t.type = 'ANGLE_CLOSE'
        return t

    # lexer.ll:405
    # <figures>_

    # lexer.ll:408
    # <figures>\>

    # lexer.ll:411
    # <figures>\<

    # lexer.ll:417
    # <notes,figures>{ALPHAWORD}
    @lex.TOKEN(ALPHAWORD)
    def t_notes_417(self, t):
        from abjad.ly import drums
        pitch_names = self.client._pitch_names
        value = t.value
        if value in pitch_names:
            t.type = 'NOTENAME_PITCH'
            t.value = pitch_names[t.value]
        elif value in drums:
            t.type = 'NOTENAME_PITCH'
            t.value = drums[value]
        elif value in ['r', 's']:
            t.type = 'RESTNAME'
        elif value == 'R':
            t.type = 'MULTI_MEASURE_REST'
        elif value == 'q':
            if self.client._last_chord is None:
                self.client._last_chord = core.Chord(['c', 'g', "c'"], (1, 4))
            t.type = 'CHORD_REPETITION'
        else:
            t.type = 'STRING'
        return t

    # lexer.ll:421
    # <notes,figures>{NOTECOMMAND}
    @lex.TOKEN(NOTECOMMAND)
    def t_notes_421(self, t):
        t.type = self.scan_escaped_word(t)
        return t

    # lexer.ll:424
    # <notes,figures>{FRACTION}
    @lex.TOKEN(FRACTION)
    def t_notes_424(self, t):
        from abjad import parser as abjad_parser
        t.type = 'FRACTION'
        parts = t.value.split('/')
        t.value = abjad_parser.LilyPondFraction(int(parts[0]), int(parts[1]))
        return t

    # lexer.ll:428
    # <notes,figures>{UNSIGNED}/\/|{UNSIGNED}
    #@lex.TOKEN('%s/\/|%s' % (UNSIGNED, UNSIGNED))
    @lex.TOKEN('%s/\/' % UNSIGNED)
    def t_notes_428(self, t):
        t.type = 'UNSIGNED'
        t.value = int(t.value)
        return t

    @lex.TOKEN(UNSIGNED)
    def t_notes_428b(self, t):
        t.type = 'UNSIGNED'
        t.value = int(t.value)
        return t

    # lexer.ll:433
    # <notes,figures>{E_UNSIGNED}
    @lex.TOKEN(E_UNSIGNED)
    def t_notes_433(self, t):
        t.type = 'E_UNSIGNED'
        t.value = int(t.value[1:])
        return t

    # lexer.ll:440
    # <quote,lyric_quote>\\{ESCAPED}
    @lex.TOKEN('\\%s' % ESCAPED)
    def t_quote_440(self, t):
        self.string_accumulator += t.value
        pass

    def t_quote_XXX(self, t):
        r'\\"'
        self.string_accumulator += t.value
        pass

    # lexer.ll:443
    # <quote,lyric_quote>[^\\""]+
    def t_quote_443(self, t):
        r'[^\\""]+'
        self.string_accumulator += t.value
        pass

    # lexer.ll:446
    # <quote,lyric_quote>\"
    def t_quote_446(self, t):
        r'\"'
        t.lexer.pop_state()
        t.type = 'STRING'
        t.value = self.string_accumulator
        return t

    # lexer.ll:456
    # <quote,lyric_quote>.
    def t_quote_456(self, t):
        r'.'
        self.string_accumulator += t.value
        pass

    # lexer.ll:462
    # <lyrics>\"

    # lexer.ll:465
    # <lyrics>{FRACTION}

    # lexer.ll:469
    # <lyrics>{UNSIGNED}/\/[^0-9]

    # lexer.ll:473
    # <lyrics>{UNSIGNED}/\/|{UNSIGNED}

    # lexer.ll:478
    # <lyrics>{NOTECOMMAND}

    # lexer.ll:481
    # <lyrics>{LYRICS}

    # lexer.ll:499
    # <lyrics>.

    # lexer.ll:504
    # <chords>{ALPHAWORD}

    # lexer.ll:507
    # <chords>{NOTECOMMAND}

    # lexer.ll:510
    # <chords>{FRACTION}

    # lexer.ll:514
    # <chords>{UNSIGNED}/\/[^0-9]

    # lexer.ll:518
    # <chords>{UNSIGNED}/\/|{UNSIGNED}

    # lexer.ll:523
    # <chords>-

    # lexer.ll:526
    # <chords>:

    # lexer.ll:529
    # <chords>\/\+

    # lexer.ll:532
    # <chords>\/

    # lexer.ll:535
    # <chords>\^

    # lexer.ll:538
    # <chords>.

    # lexer.ll:545
    # <markup>\\score
    def t_markup_545(self, t):
        r'\\score'
        t.type = 'SCORE'
        return t

    # lexer.ll:548
    # <markup>{MARKUPCOMMAND}
    @lex.TOKEN(MARKUPCOMMAND)
    def t_markup_548(self, t):
        value = t.value[1:]
        if value in self.client._markup_functions or \
            value in self.client._markup_list_functions:
            if value in self.client._markup_functions:
                t.type = 'MARKUP_FUNCTION'
                signature = self.client._markup_functions[value]
            else:
                t.type = 'MARKUP_LIST_FUNCTION'
                signature = self.client._markup_list_functions[value]
            #print t.type, value, signature
            self.push_signature(signature, t)
        else:
            t.type = self.scan_escaped_word(t)
        return t

    # lexer.ll:598
    # <markup>[{}]
#    def t_markup_598(self, t):
#        r'[{}]'
#        t.type = t.value
#        return t

    # lexer.ll:601
    # <markup>[^#{}\"\\ \t\n\r\f]+
    def t_markup_601(self, t):
        r'[^#{}\"\\ \t\n\r\f]+'
        t.type = 'STRING'
        return t

    # lexer.ll:614
    # <markup>.

    # lexer.ll:619
    # <longcomment><<EOF>>

    # lexer.ll:626
    # <<EOF>>

    # lexer.ll:643
    # <INITIAL>{DASHED_WORD}
    @lex.TOKEN(DASHED_WORD)
    def t_INITIAL_643(self, t):
        t.type = self.scan_bare_word(t)
        return t

    # lexer.ll:646
    # <INITIAL>{DASHED_KEY_WORD}
    @lex.TOKEN(DASHED_KEY_WORD)
    def t_INITIAL_646(self, t):
        t.type = self.scan_escaped_word(t)
        return t

    # lexer.ll:651
    # -{UNSIGNED}|{REAL}
    @lex.TOKEN(REAL)
    def t_651_a(self, t):
        t.type = 'REAL'
        t.value = float(t.value)
        return t

    @lex.TOKEN('-%s' % UNSIGNED)
    def t_651_b(self, t):
        t.type = 'REAL'
        t.value = float(t.value)
        return t

    # lexer.ll:661
    # -\.
    def t_661(self, t):
        '-\.'
        t.type = 'REAL'
        t.value = 0.0
        return t

    # lexer.ll:666
    # {UNSIGNED}
    @lex.TOKEN(UNSIGNED)
    def t_666(self, t):
        t.type = 'UNSIGNED'
        t.value = float(t.value)
        return t

    # lexer.ll:672
    # [{}]

    # lexer.ll:676
    # [*:=]

    # lexer.ll:682
    # <INITIAL,notes,figures>.

    # lexer.ll:686
    # <INITIAL,lyrics,notes,figures>\\.
    def t_INITIAL_notes_686(self, t):
        r'\\.'
        if t.value[1] == '>':
            t.type = 'E_ANGLE_CLOSE'
        elif t.value[1] == '<':
            t.type = 'E_ANGLE_OPEN'
        elif t.value[1] == '!':
            t.type = 'E_EXCLAMATION'
        elif t.value[1] == '(':
            t.type = 'E_OPEN'
        elif t.value[1] == ')':
            t.type = 'E_CLOSE'
        elif t.value[1] == '[':
            t.type = 'E_BRACKET_OPEN'
        elif t.value[1] == '+':
            t.type = 'E_PLUS'
        elif t.value[1] == ']':
            t.type = 'E_BRACKET_CLOSE'
        elif t.value[1] == '~':
            t.type = 'E_TILDE'
        elif t.value[1] == '\\':
            t.type = 'E_BACKSLASH'
        else:
            t.type = 'E_CHAR'
        return t

    # lexer.ll:
    # <*>.716
#    def t_ANY_716(self, t):
#        r'.'
#        raise Exception

    ### DEFAULT RULES ###

    t_ignore = '' # let the grammar handle ignoring things

    #    t_extratoken_ignore = t_ignore
    #    t_chords_ignore = t_ignore
    #    t_figures_ignore = t_ignore
    #    t_incl_ignore = t_ignore
    #    t_lyrics_ignore = t_ignore
    #    t_lyric_quote_ignore = t_ignore
    t_longcomment_ignore = t_ignore
    t_markup_ignore = t_ignore
    t_notes_ignore = t_ignore
    t_quote_ignore = t_ignore
    #    t_sourcefileline_ignore = t_ignore
    #    t_sourcefilename_ignore = t_ignore
    t_version_ignore = t_ignore

    t_scheme_ignore = t_ignore

    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count("\n")

    def t_error(self, t):
        print(("LilyPondParser: Illegal character '%s'" % t.value[0]))
        t.lexer.skip(1)

    #    t_extratoken_error = t_error
    #    t_chords_error = t_error
    #    t_figures_error = t_error
    #    t_incl_error = t_error
    #    t_lyrics_error = t_error
    #    t_lyric_quote_error = t_error
    t_longcomment_error = t_error
    t_markup_error = t_error
    t_notes_error = t_error
    t_quote_error = t_error
    #    t_sourcefileline_error = t_error
    #    t_sourcefilename_error = t_error
    t_version_error = t_error
    t_scheme_error = t_error

    def scan_bare_word(self, t):
        if t.lexer.current_state() in ('notes',):
            pitch_names = self.client._pitch_names
            if t.value in pitch_names:
                t.type = 'NOTENAME_PITCH'
            elif t.value == 'q' and self.client._last_chord:
                t.type = 'CHORD_REPETITION'
        return 'STRING'

    def scan_escaped_word(self, t):
        from abjad import parser as abjad_parser
        # first, check for it in the keyword list
        if t.value in self.keywords:
            value = self.keywords[t.value]
            if value == 'MARKUP':
                t.lexer.push_state('markup')
                if t.lexer.current_state() == 'lyrics':
                    return 'LYRIC_MARKUP'
            elif value == 'WITH':
                t.lexer.push_state('INITIAL')
            return value
        identifier = t.value[1:]
        # check for the identifier in the scope stack
        lookup = self.client._resolve_identifier(identifier)
        if lookup is not None:
            identifier_lookup = {
                'book_block': 'BOOK_IDENTIFIER',
                'bookpart_block': 'BOOK_IDENTIFIER',
                'context_def_spec_block': 'CONTEXT_DEF_IDENTIFIER',
                'context_modification': 'CONTEXT_MOD_IDENTIFIER',
                'post_event_nofinger': 'EVENT_IDENTIFIER',
                'full_markup': 'MARKUP_IDENTIFIER',
                'full_markup_list': 'MARKUPLINES_IDENTIFIER',
                'music': 'MUSIC_IDENTIFIER',
                'number_expression': 'NUMBER_IDENTIFIER',
                'output_def': 'output_DEF_IDENTIFIER',
                'embedded_scm': 'SCM_IDENTIFIER',
                'score_block': 'SCORE_IDENTIFIER',
                'string': 'STRING_IDENTIFIER',
                # 'PITCH_IDENTIFIER' ?
                # 'DURATION_IDENTIFIER' ?
                # 'LYRIC_MARKUP_IDENTIFIER' ?
            }
            t.value = copy.deepcopy(lookup.value)
            return identifier_lookup[lookup.type]
        # then, check for it in the "current_module" dictionary
        # which we've dumped out of LilyPond
        if identifier not in self.client._current_module:
            message = 'unknown escaped word: {!r}.'
            message = message.format(t.value)
            raise Exception(message)
        lookup = self.client._current_module[identifier]
        # if the lookup resolves to a function definition,
        # we have to push artificial tokens onto the token stack.
        # the tokens are pushed in reverse order (LIFO).
        if isinstance(lookup, dict) and 'type' in lookup:
            if lookup['type'] == 'ly:music-function?':
                signature = lookup['signature']
                funtype = 'SCM_FUNCTION'
                if signature[0] == 'ly:music?':
                    funtype = 'MUSIC_FUNCTION'
                elif signature[0] == 'ly:event?':
                    funtype = 'EVENT_FUNCTION'
                self.push_signature(signature[1:], t)
                return funtype
            elif lookup['type'] == 'ly:prob?':
                if 'event' in lookup['types']:
                    return 'EVENT_IDENTIFIER'
                elif (
                    'context-specification' in lookup['types'] and
                    hasattr(self.client._guile, identifier)
                    ):
                    t.value = getattr(self.client._guile, identifier)()
                    return 'MUSIC_IDENTIFIER'
        # we also check for other types, to handle \longa, \breve etc.
        elif isinstance(lookup, abjad_parser.LilyPondDuration):
            t.value = copy.copy(lookup)
            return 'DURATION_IDENTIFIER'
        # else...
        t.value = copy.copy(lookup)
        return 'SCM_IDENTIFIER'

    def push_signature(self, signature, t):

        token = lex.LexToken()
        token.type = 'EXPECT_NO_MORE_ARGS'
        token.value = None
        token.lineno = t.lineno
        token.lexpos = t.lexpos
        self.client._push_extra_token(token)

        optional = False
        for predicate in signature:

            if predicate == 'optional?':
                optional = True
                continue

            token = lex.LexToken()
            token.value = predicate
            token.lineno = t.lineno
            token.lexpos = t.lexpos

            if predicate == 'ly:music?':
                token.type = 'EXPECT_SCM' # ?!?!
            elif predicate == 'ly:pitch?':
                token.type = 'EXPECT_PITCH'
            elif predicate == 'ly:duration?':
                token.type = 'EXPECT_DURATION'
            elif predicate in ['markup?', 'cheap-markup?']:
                token.type = 'EXPECT_MARKUP'
            elif predicate == 'markup-list?':
                token.type = 'EXPECT_MARKUP_LIST'
            else:
                token.type = 'EXPECT_SCM'

            self.client._push_extra_token(token)

            if optional:
                optional_token = lex.LexToken()
                optional_token.value = 'optional?'
                optional_token.lineno = t.lineno
                optional_token.lexpos = t.lexpos
                optional_token.type = 'EXPECT_OPTIONAL'
                self.client._push_extra_token(optional_token)
                optional = False



import typing
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.lilypondnames.LilyPondTweakManager import LilyPondTweakManager
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle


class Fermata(AbjadValueObject):
    r"""
    Fermata.

    ..  container:: example

        A short fermata:

        >>> score = abjad.Score([abjad.Staff([abjad.Note("c'4")])])
        >>> fermata = abjad.Fermata(command='shortfermata')
        >>> abjad.attach(fermata, score[0][0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    c'4
                    \shortfermata
                }
            >>

    ..  container:: example

        A fermata:

        >>> score = abjad.Score([abjad.Staff([abjad.Note("c'4")])])
        >>> fermata = abjad.Fermata()
        >>> abjad.attach(fermata, score[0][0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    c'4
                    \fermata
                }
            >>

    ..  container:: example

        A long fermata:

        >>> score = abjad.Score([abjad.Staff([abjad.Note("c'4")])])
        >>> fermata = abjad.Fermata('longfermata')
        >>> abjad.attach(fermata, score[0][0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    c'4
                    \longfermata
                }
            >>

    ..  container:: example

        A very long fermata:

        >>> score = abjad.Score([abjad.Staff([abjad.Note("c'4")])])
        >>> fermata = abjad.Fermata('verylongfermata')
        >>> abjad.attach(fermata, score[0][0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    c'4
                    \verylongfermata
                }
            >>

    """

    ### CLASS VARIABLES ###

    _allowable_commands = (
        'fermata',
        'longfermata',
        'shortfermata',
        'verylongfermata',
        )

    __slots__ = (
        '_command',
        '_tweaks',
        )

    _context = 'Score'

    _format_slot = 'after'

    ### INITIALIZER ###

    def __init__(
        self,
        command: str = 'fermata',
        *,
        tweaks: typing.Union[
            typing.List[typing.Tuple], LilyPondTweakManager] = None,
        ) -> None:
        assert command in self._allowable_commands, repr(command)
        self._command = command
        self._tweaks = None
        LilyPondTweakManager.set_tweaks(self, tweaks)

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        r"""
        Gets string representation of fermata.

        ..  container:: example

            Fermata:

            >>> str(abjad.Fermata())
            '\\fermata'

        ..  container:: example

            Long fermata:

            >>> str(abjad.Fermata('longfermata'))
            '\\longfermata'

        Returns string.
        """
        return rf'\{self.command}'

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        return str(self)

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if self.tweaks:
            tweaks = self.tweaks._list_format_contributions()
            bundle.after.articulations.extend(tweaks)
        bundle.after.articulations.append(self._get_lilypond_format())
        return bundle

    ### PUBLIC METHODS ###

    @staticmethod
    def list_allowable_commands() -> typing.Tuple[str, ...]:
        """
        Lists allowable commands:

        ..  container:: example

            All allowable commands:

            >>> commands = abjad.Fermata.list_allowable_commands()
            >>> for command in commands:
            ...     command
            'fermata'
            'longfermata'
            'shortfermata'
            'verylongfermata'

        """
        return Fermata._allowable_commands

    ### PUBLIC PROPERTIES ###

    @property
    def command(self) -> typing.Optional[str]:
        """
        Gets command of fermata.

        ..  container:: example

            Fermata:

            >>> fermata = abjad.Fermata()
            >>> fermata.command
            'fermata'

        ..  container:: example

            Long fermata:

            >>> fermata = abjad.Fermata('longfermata')
            >>> fermata.command
            'longfermata'

        """
        return self._command

    @property
    def context(self) -> str:
        """
        Gets (historically conventional) context.

        ..  container:: example

            Fermata:

            >>> fermata = abjad.Fermata()
            >>> fermata.context
            'Score'

        ..  container:: example

            Long fermata:

            >>> fermata = abjad.Fermata('longfermata')
            >>> fermata.context
            'Score'

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def tweaks(self) -> typing.Optional[LilyPondTweakManager]:
        r"""
        Gets tweaks

        ..  container:: example

            >>> note = abjad.Note("c'4")
            >>> fermata = abjad.Fermata()
            >>> abjad.tweak(fermata).color = 'blue'
            >>> abjad.attach(fermata, note)
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                c'4
                - \tweak color #blue
                \fermata

        ..  container:: example

            >>> note = abjad.Note("c'4")
            >>> fermata = abjad.Fermata(tweaks=[('color', 'blue')])
            >>> abjad.attach(fermata, note)
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                c'4
                - \tweak color #blue
                \fermata

        """
        return self._tweaks
